# -*- coding: utf-8 -*-
"""
Reads an `html` formatted table.
"""


import numpy as np
from bs4 import BeautifulSoup
import requests
from selenium import webdriver
from selenium.webdriver.firefox.options import Options as FirefoxOptions
from selenium.webdriver.chrome.options import Options as ChromeOptions
from selenium.webdriver.edge.options import Options as EdgeOptions
from selenium.webdriver.ie.options import Options as IeOptions
import copy
import logging
from tabledataextractor.exceptions import InputError

log = logging.getLogger(__name__)


def makearray(html_table):
    """
    Creates a numpy array from an `.html` file, taking `rowspan` and `colspan` into account.

    Modified from:
        John Ricco, https://johnricco.github.io/2017/04/04/python-html/, *Using Python to scrape HTML tables with merged cells*

    Added functionality for duplicating cell content for cells with `rowspan`/`colspan`.
    The table has to be :math:`n*m`, rectangular, with the same number of columns in every row.
    """
    n_cols = 0
    n_rows = 0

    for row in html_table.findAll("tr"):
        col_tags = row.find_all(["td", "th"])
        if len(col_tags) > 0:
            n_rows += 1
            if len(col_tags) > n_cols:
                n_cols = len(col_tags)

    # according to numpy documentation fill_value should be of type Union[int, float, complex]
    # however, 'str' works just fine
    array = np.full((n_rows, n_cols), fill_value="", dtype='<U60')

    # list to store rowspan values
    skip_index = [0 for i in range(0, n_cols)]

    # iterating over each row in the table
    row_counter = 0
    for row in html_table.findAll("tr"):

        # skip row if it's empty
        if len(row.find_all(["td", "th"])) == 0:
            continue

        else:

            # get all the cells containing data in this row
            columns = row.find_all(["td", "th"])
            col_dim = []
            row_dim = []
            col_dim_counter = -1
            row_dim_counter = -1
            col_counter = -1
            this_skip_index = copy.deepcopy(skip_index)

            for col in columns:

                # determine all cell dimensions
                colspan = col.get("colspan")
                if not colspan:
                    col_dim.append(1)
                else:
                    col_dim.append(int(colspan))
                col_dim_counter += 1

                rowspan = col.get("rowspan")
                if not rowspan:
                    row_dim.append(1)
                else:
                    row_dim.append(int(rowspan))
                row_dim_counter += 1

                # adjust column counter
                if col_counter == -1:
                    col_counter = 0
                else:
                    col_counter = col_counter + col_dim[col_dim_counter - 1]

                while skip_index[col_counter] > 0:
                    col_counter += 1

                # get cell contents
                cell_data = col.get_text()

                # insert data into cell
                array[row_counter, col_counter] = cell_data

                # Insert data into neighbouring rowspan/colspan cells
                if colspan:
                    for spanned_col in range(col_counter+1, col_counter + int(colspan)):
                        array[row_counter, spanned_col] = cell_data
                if rowspan:
                    for spanned_row in range(row_counter+1, row_counter + int(rowspan)):
                        array[spanned_row, col_counter] = cell_data

                #record column skipping index
                if row_dim[row_dim_counter] > 1:
                    this_skip_index[col_counter] = row_dim[row_dim_counter]

        # adjust row counter
        row_counter += 1

        # adjust column skipping index
        skip_index = [i - 1 if i > 0 else i for i in this_skip_index]

    return array


def read_file(file_path, table_number=1):
    """Reads an .html file and returns a numpy array."""
    file = open(file_path, encoding='UTF-8')
    html_soup = BeautifulSoup(file, features='lxml')
    file.close()
    html_table = html_soup.find_all("table")[table_number-1]
    array = makearray(html_table)
    return array


def configure_selenium(browser='Firefox'):
    """
    Configuration for `Selenium <https://selenium-python.readthedocs.io/>`_. Sets the path to ``geckodriver.exe``

    :param browser: Which browser to use
    :type browser: str
    :return: Selenium driver

    """
    if browser == 'Firefox':
        options = FirefoxOptions()
        options.headless = True
        driver = webdriver.Firefox(options=options, executable_path=r'C:\Users\juras\System\geckodriver\geckodriver.exe')
        return driver
    else:
        return None


def read_url(url, table_number=1):
    """
    Reads in a table from an URL and returns a numpy array. Will try `Requests <http://docs.python-requests.org/en/master/>`_ first. If it doesn't succeed, `Selenium <https://selenium-python.readthedocs.io/>`_ will be used.

    :param url: Url of the page where the table is located
    :type url: str
    :param table_number: Number of Table on the web page.
    :type table_number: int
    """

    if not isinstance(table_number, int):
        msg = 'Table number is not valid.'
        log.critical(msg)
        raise TypeError(msg)

    # first try the requests package, if it fails do the selenium, which is much slower
    try:
        html_file = requests.get(url)
        html_soup = BeautifulSoup(html_file.text, features='lxml')
        html_table = html_soup.find_all("table")[table_number - 1]
        array = makearray(html_table)
        log.info("Package 'requests' was used.")
        return array
    except Exception:
        driver = configure_selenium()
        driver.get(url)
        html_file = driver.page_source
        html_soup = BeautifulSoup(html_file, features='lxml')
        try:
            html_table = html_soup.find_all("table")[table_number-1]
        except IndexError:
            raise InputError("table_number={} is out of range".format(table_number))
        else:
            array = makearray(html_table)
            log.info("Package 'selenium' was used.")
            return array



# -*- coding: utf-8 -*-
"""
Inputs from python list object.
"""

import numpy as np
import logging

log = logging.getLogger(__name__)


def read(plist):
    """
    Creates a numpy array from a Python list. Works if rows are of different length.

    :param plist: Input List
    :type plist: list
    :return: numpy.ndarray
    """
    length = len(sorted(plist, key=len, reverse=True)[0])
    array = np.array([l+[None]*(length-len(l)) for l in plist], dtype='<U60')
    return array




# -*- coding: utf-8 -*-
"""
Analyzes the input and calls the appropriate input module.
"""

from django.core.validators import URLValidator
from django.core.exceptions import ValidationError
import os.path
from tabledataextractor.input import from_html
from tabledataextractor.input import from_csv
from tabledataextractor.input import from_list
import logging

log = logging.getLogger(__name__)


def url(name):
    """
    Returns `True` if input is `URL`.
    Uses ``django.core.validators.URLValidator``.

    :param name: Input string
    :type name: str
    """
    try:
        URLValidator()(name)
        return True
    except ValidationError:
        return False


def html(name):
    """
    Returns `True` if input is `html` file.

    :param name: Input string
    :type name: str
    """
    if os.path.isfile(name) and name.endswith(".html"):
        return True
    else:
        return False


def csv(name):
    """
    Returns `True` if input is `csv` file.

    :param name: Input string
    :type name: str
    """
    if os.path.isfile(name) and name.endswith(".csv"):
        return True
    else:
        return False


def create_table(name_key, table_number=1):
    """
    Checks the input and calls the appropriate modules for conversion.
    Returns a numpy array with the raw table.

    :param name_key: Path to `.html` or `.cvs` file, `URL` or `python list` that is used as input
    :type name_key: str | list
    :param table_number: Number of the table that we want to input if there are several at the given address/path
    :type table_number: int
    :return: table as numpy.array
    """

    if isinstance(name_key, list):
        log.info("Input is list type.")
        if len(name_key) > 0:
            return from_list.read(name_key)
        else:
            msg = 'Input is invalid. ' \
                  'Supported are: path to .html or .cvs file, URL or multidimensional python list object'
            log.critical(msg)
            raise TypeError(msg, str(name_key))

    elif url(name_key):
        log.info("Url: {}".format(name_key))
        return from_html.read_url(name_key, table_number)

    elif html(name_key):
        log.info("HTML File: {}".format(name_key))
        return from_html.read_file(name_key, table_number)

    elif csv(name_key):
        log.info("CSV File: {}".format(name_key))
        return from_csv.read(name_key)

    else:
        msg = 'Input is invalid. Supported are: path to .html or .cvs file, URL or multidimensional python list object'
        log.critical(msg)
        raise TypeError(msg, str(name_key))





# -*- coding: utf-8 -*-
"""
Algorithms for TableDataExtractor.

.. codeauthor:: Juraj Mavračić <jm2111@cam.ac.uk>

"""

import logging
import numpy as np
from sympy import Symbol
from sympy import factor_list, factor

from tabledataextractor.exceptions import MIPSError
from tabledataextractor.table.parse import StringParser, CellParser


log = logging.getLogger(__name__)


def empty_string(string, regex=r'^([\s\-\–\—\"]+)?$'):
    """
    Returns `True` if a particular string is empty, which is defined with a regular expression.

    :param string: Input string for testing
    :type string: str
    :param regex: The regular expression which defines an empty cell (can be tweaked).
    :type regex: str
    :return: True/False
    """
    empty_parser = StringParser(regex)
    return empty_parser.parse(string, method='fullmatch')


def empty_cells(array, regex=r'^([\s\-\–\—\"]+)?$'):
    """
    Returns a mask with `True` for all empty cells in the original array and `False` for non-empty cells.

    :param regex: The regular expression which defines an empty cell (can be tweaked).
    :type regex: str
    :param array: Input array to return the mask for
    :type array: numpy array
    """
    empty = np.full_like(array, fill_value=False, dtype=bool)
    empty_parser = CellParser(regex)
    for empty_cell in empty_parser.parse(array, method='fullmatch'):
        if array.ndim == 2:
            empty[empty_cell[0], empty_cell[1]] = True
        elif array.ndim == 1:
            empty[empty_cell[0]] = True
    return empty


def standardize_empty(array):
    """
    Returns an array with the empty cells of the input array standardized to 'NoValue'.

    :param array: Input array
    :type array: numpy.array
    :return: Array with standardized empty cells
    """
    standardized = np.copy(array)
    for row_index, row in enumerate(standardized):
        for col_index, col in enumerate(row):
            if empty_string(col):
                standardized[row_index, col_index] = 'NoValue'
    return standardized


def pre_clean(array):
    """
    Removes empty and duplicate rows and columns that extend over the whole table.

    :param array: Input Table object
    :type array: Numpy array
    """

    pre_cleaned_table = np.copy(array)
    array_empty = empty_cells(array)

    # find empty rows and delete them
    empty_rows = []
    for row_index, row in enumerate(array_empty):
        if False not in row:
            empty_rows.append(row_index)
    log.debug("Empty rows {} deleted.".format(empty_rows))
    pre_cleaned_table = np.delete(pre_cleaned_table, empty_rows, axis=0)

    # find empty columns and delete them
    empty_columns = []
    for column_index, column in enumerate(array_empty.T):
        if False not in column:
            empty_columns.append(column_index)
    log.debug("Empty columns {} deleted.".format(empty_columns))
    pre_cleaned_table = np.delete(pre_cleaned_table, empty_columns, axis=1)

    # delete duplicate rows that extend over the whole table
    _, indices = np.unique(pre_cleaned_table, axis=0, return_index=True)
    # for logging only, which rows have been removed
    removed_rows = []
    for row_index in range(0, len(pre_cleaned_table)):
        if row_index not in indices:
            removed_rows.append(row_index)
    log.debug("Duplicate rows {} removed.".format(removed_rows))
    # deletion:
    pre_cleaned_table = pre_cleaned_table[np.sort(indices)]

    # delete duplicate columns that extend over the whole table
    _, indices = np.unique(pre_cleaned_table, axis=1, return_index=True)
    # for logging only, which rows have been removed
    removed_columns = []
    for column_index in range(0, len(pre_cleaned_table.T)):
        if column_index not in indices:
            removed_columns.append(column_index)
    log.debug("Duplicate columns {} removed.".format(removed_columns))
    # deletion:
    pre_cleaned_table = pre_cleaned_table[:, np.sort(indices)]

    # clean-up unicode characters
    pre_cleaned_table = clean_unicode(pre_cleaned_table)

    return pre_cleaned_table


def clean_unicode(array):
    """
    Replaces problematic unicode characters in a given numpy array.
    :param array: input array
    :type array: numpy.array
    :return: cleaned array
    """
    temp = np.copy(array)
    temp = np.char.replace(temp, '\xa0', ' ')
    return temp


def find_cc4(table_object):
    """
    Searches for critical cell `CC4`.

    Searching from the bottom of the pre-cleaned table for the last row with a minority of empty cells.
    Rows with at most a few empty cells are assumed to be part of the data region rather than notes or footnotes rows
    (which usually only have one or two non-empty cells).

    :param table_object: Input Table object
    :type table_object: ~tabledataextractor.table.table.Table
    :return: cc4
    """
    # searching from the bottom of original table:
    n_rows = len(table_object.pre_cleaned_table)
    for row_index in range(n_rows - 1, -1, -1):
        # counting the number of full cells
        # if n_empty < n_full terminate, this is our goal row
        n_full = 0
        n_columns = len(table_object.pre_cleaned_table_empty[row_index])
        for empty in table_object.pre_cleaned_table_empty[row_index]:
            if not empty:
                n_full += 1
            if n_full > int(n_columns / 2):
                return row_index, n_columns - 1


def duplicate_rows(table):
    """
    Returns True if there are duplicate rows in the table and False if there are no duplicate rows
    :param table:
    :return: True or False
    """
    if table.ndim > 0 and table.size:
        _, indices = np.unique(table, axis=0, return_index=True)
        if len(table) > len(indices):
            return True
        else:
            return False
    else:
        return False


def duplicate_columns(table):
    """
    Returns True if there are duplicate columns in the table and False if there are no duplicate columns
    :param table:
    :return: True or False
    """
    if table.T.ndim > 0 and table.T.size:
        _, indices = np.unique(table.T, axis=0, return_index=True)
        if len(table.T) > len(indices):
            return True
        else:
            return False
    else:
        return False


def find_cc1_cc2(table_object, cc4, array):
    """
    Main MIPS (*Minimum Indexing Point Search*) algorithm. According to Embley et al., *DOI: 10.1007/s10032-016-0259-1*.
    Searches for critical cells `CC2` and `CC3`.
    MIPS locates the critical cells that define the minimum row and column headers needed to index
    every data cell.

    :param table_object: Input Table object
    :type table_object: ~tabledataextractor.table.table.Table
    :param cc4: Position of `CC4` cell found with ``find_cc4()``
    :param array: table to search for `CC1` and `CC2`
    :type array: numpy array
    :type cc4: (int, int)
    :return: cc1, cc2
    """

    # Initialize
    cc2 = None
    c_max = cc4[1]
    r_max = cc4[0]
    r1 = 0
    c1 = 0
    r2 = r_max - 1
    c2 = 0
    max_area = 0

    def table_slice_cc2(table, r2, r_max, c1, c2):
        """
        Function to cut the correct slices out of array for `CC2 `in ``find_cc1_cc2()``.
        Cuts out the next row and column header candidates from the pre-cleaned table.

        :param table: pre-cleaned table
        :param r2: current r2 parameter in MIPS algorithm
        :param r_max: r_max parameter in MIPS algorithm
        :param c1: first column for MIPS algorithm
        :param c2: current c2 parameter for MIPS algorithm
        :return: (section_1, section_2)
        """

        # one more row and column index than in the published pseudocode is needed,
        # since the a:b notation in python doesn't include b
        if r2 + 1 == r_max and c1 == c2:
            section_1 = table[r2 + 1, c1]
        elif r2 + 1 == r_max and c1 != c2:
            section_1 = table[r2 + 1, c1:c2 + 1]
        elif r2 + 1 != r_max and c1 != c2:
            section_1 = table[r2 + 1:r_max + 1, c1:c2 + 1]
        elif r2 + 1 != r_max and c1 == c2:
            section_1 = table[r2 + 1:r_max + 1, c1]
        else:
            log.critical("Not defined section_1, r2+1= {}, r_max= {}, c1= {}, c2= {}".format(r2 + 1, r_max, c1, c2))
            section_1 = None

        # contrary to the published pseudocode the row maximum is r2, not r2-1
        # one more row and column index than in the published pseudocode is needed,
        # since the a:b notation in python doesn't include b
        if r1 == r2 and c2 + 1 == c_max:
            section_2 = table[r1, c2 + 1]
        elif r1 == r2 and c2 + 1 != c_max:
            section_2 = table[r1, c2 + 1: c_max + 1]
        elif r1 != r2 and c2 + 1 != c_max:
            section_2 = table[r1: r2 + 1, c2 + 1: c_max + 1]
        elif r1 != r2 and c2 + 1 == c_max:
            section_2 = table[r1: r2 + 1, c2 + 1]
        else:
            log.critical(
                "Not defined section_2, r2-1= {}, r1= {}, c2+1= {}, c_max= {}".format(r2 - 1, r1, c2 + 1, c_max))
            section_2 = None

        return section_1, section_2

    def table_slice_1_cc1(table, r1, r2, c2, c_max):
        """
        Function to cut a correct slice out of array for CC1 in _find_cc1_cc2().
        Cuts out the column header.
        """
        # one more row and column index than in the published pseudocode is needed,
        # since the a:b notation in python doesn't include b
        # contrary to the published pseudocode, the correct range is [r1:r2,c2+1:c_max] and not [r1+1:r2,c2+1:c_max]
        if r1 == r2 and c2 + 1 == c_max:
            section = table[r1, c2 + 1]
        elif r1 == r2 and c2 + 1 != c_max:
            section = table[r1, c2 + 1:c_max + 1]
        elif r1 != r2 and c2 + 1 != c_max:
            section = table[r1: r2 + 1, c2 + 1:c_max + 1]
        elif r1 != r2 and c2 + 1 == c_max:
            section = table[r1: r2 + 1, c2 + 1]
        else:
            log.critical(
                "Not defined section 1 for cc1, r1+1= {}, r2= {}, c2+1= {}, c_max= {}".format(r1 + 1, r2, c2 + 1,
                                                                                              c_max))
            section = None
        return section

    def table_slice_2_cc1(table, r2, r_max, c1, c2):
        """
        Function to cut a correct slice out of array for CC1 in _find_cc1_cc2().
        Cuts out the row header.
        """
        # one more row and column index than in the published pseudocode is needed,
        # since the a:b notation in python doesn't include b
        # contrary to the published pseudocode, the correct range is [r2:r_max,c1:c2] and not [r2+1:c2,c1+1:r_max]
        if r2 + 1 == r_max and c1 == c2:
            section = table[r2 + 1, c1]
        elif r2 + 1 == r_max and c1 != c2:
            section = table[r2 + 1, c1: c2 + 1]
        elif r2 + 1 != r_max and c1 != c2:
            section = table[r2 + 1: r_max + 1, c1: c2 + 1]
        elif r2 + 1 != r_max and c1 == c2:
            section = table[r2 + 1: r_max + 1, c1]
        else:
            log.critical(
                "Not defined section 2 for cc1, r2+1= {}, c2= {}, c1+1= {}, r_max= {}".format(r2 + 1, c2, c1 + 1,
                                                                                              r_max))
            section = None
        return section

    # MAIN MIPS algorithm
    # Locate candidate MIPs by finding the minimum indexing headers:
    # This is significantly altered compared to the published pseudocode, which is flawed.
    # The pseudocode clearly does not return cc2 if the column has not been changed and it doesn't
    # discriminate between duplicate rows in the row header vs duplicate columns in the column header
    while c2 < c_max and r2 >= r1:

        log.debug("Entering loop:  r_max= {}, c_max= {}, c1= {}, c2= {}, r1= {}, r2= {}, cc2= {}"
                  .format(r_max, c_max, c1, c2, r1, r2, cc2))

        temp_section_1, temp_section_2 = table_slice_cc2(array, r2, r_max, c1, c2)

        log.debug("temp_section_1:\n{}".format(temp_section_1))
        log.debug("temp_section_2:\n{}".format(temp_section_2))
        log.debug("duplicate_rows= {}, duplicate_columns= {}".
                  format(duplicate_rows(temp_section_1), duplicate_rows(temp_section_2)))

        if not duplicate_rows(temp_section_1) and not duplicate_columns(temp_section_2):
            if table_object.configs['use_max_data_area']:
                data_area = (r_max - r2) * (c_max - c2)
                log.debug("The data area of the new candidate C2= {} is *1: {}".format((r2, c2), data_area))
                log.debug("Data area:\n{}".format(array[r2 + 1:r_max + 1, c2 + 1:c_max + 1]))
                if data_area >= max_area:
                    max_area = data_area
                    cc2 = (r2, c2)
                    log.debug("CC2= {}".format(cc2))
                r2 = r2 - 1
            else:
                cc2 = (r2, c2)
                log.debug("CC2= {}".format(cc2))
                r2 = r2 - 1
        elif duplicate_rows(temp_section_1) and not duplicate_columns(temp_section_2):
            c2 = c2 + 1
            if table_object.configs['use_max_data_area']:
                data_area = (r_max - r2) * (c_max - c2)
                log.debug("The data area of the new candidate C2= {} is *2: {}".format((r2, c2), data_area))
                log.debug("Data area:\n{}".format(array[r2 + 1:r_max + 1, c2 + 1:c_max + 1]))
                if data_area >= max_area:
                    max_area = data_area
                    cc2 = (r2, c2)
                    log.debug("CC2= {}".format(cc2))
            else:
                cc2 = (r2, c2)
                log.debug("CC2= {}".format(cc2))
        elif duplicate_rows(temp_section_1) and duplicate_columns(temp_section_2):
            c2 = c2 + 1
            r2 = r2 + 1
            if table_object.configs['use_max_data_area']:
                data_area = (r_max - r2) * (c_max - c2)
                log.debug("The data area of the new candidate C2= {} is *3: {}".format((r2, c2), data_area))
                log.debug("Data area:\n{}".format(array[r2 + 1:r_max + 1, c2 + 1:c_max + 1]))
                if data_area >= max_area:
                    max_area = data_area
                    cc2 = (r2, c2)
                    log.debug("CC2= {}".format(cc2))
            else:
                cc2 = (r2, c2)
        # if none of those above is satisfied, just finish the loop
        else:
            r2 = r2 + 1
            if table_object.configs['use_max_data_area']:
                data_area = (r_max - r2) * (c_max - c2)
                log.debug("The data area of the new candidate C2= {} is *4: {}".format((r2, c2), data_area))
                log.debug("Data area:\n{}".format(array[r2 + 1:r_max + 1, c2 + 1:c_max + 1]))
                if data_area >= max_area:
                    max_area = data_area
                    cc2 = (r2, c2)
                    log.debug("CC2= {}".format(cc2))
                break
            else:
                cc2 = (r2, c2)
                break

    log.debug(
        "Ended loop with:  r_max= {}, c_max= {}, c1= {}, c2= {}, r1= {}, r2= {}, cc2= {}\n\n\n\n".format(r_max,
                                                                                                         c_max, c1,
                                                                                                         c2, r1, r2,
                                                                                                         cc2))

    # re-initialization of r2 and c2 from cc2; missing in the pseudocode
    r2 = cc2[0]
    c2 = cc2[1]

    # Locate CC1 at intersection of the top row and the leftmost column necessary for indexing:
    log.debug("Potentially duplicate columns:\n{}".format(table_slice_1_cc1(array, r1, r2, c2, c_max)))
    while not duplicate_columns(table_slice_1_cc1(array, r1, r2, c2, c_max)) and r1 <= r2:
        log.debug("Potentially duplicate columns:\n{}".format(table_slice_1_cc1(array, r1, r2, c2, c_max)))
        log.debug("Duplicate columns= {}".format(duplicate_columns(table_slice_1_cc1(array, r1, r2, c2, c_max))))
        r1 = r1 + 1
        log.debug("r1= {}".format(r1))

    log.debug("Potentially duplicate rows:\n{}".format(table_slice_2_cc1(array, r2, r_max, c1, c2)))
    while not duplicate_rows(table_slice_2_cc1(array, r2, r_max, c1, c2)) and c1 <= c2:
        log.debug("Potentially duplicate rows:\n{}".format(table_slice_2_cc1(array, r2, r_max, c1, c2)))
        log.debug("Duplicate rows= {}".format(duplicate_rows(table_slice_2_cc1(array, r2, r_max, c1, c2))))
        c1 = c1 + 1
        log.debug("c1= {}".format(c1))

    # final cc1 is (r1-1,c1-1), because the last run of the while loops doesn't count
    # a problem could arise if the code never stepped through the while loops,
    # returning a cc1 with a negative index.
    # however, this should never happen since the final headers CANNOT have duplicate rows/columns,
    # by definition of cc2.
    # hence, the assertions:
    try:
        assert not duplicate_columns(table_slice_1_cc1(array, r1=0, r2=cc2[0], c2=cc2[1], c_max=c_max))
        assert not duplicate_rows(table_slice_2_cc1(array, r2=cc2[0], r_max=r_max, c1=0, c2=cc2[1]))
        assert r1 >= 0 and c1 >= 0
        cc1 = (r1 - 1, c1 - 1)
    except AssertionError:
        raise MIPSError("Error in _find_cc1_cc2")

    # provision for using the uppermost row possible for cc1, if titles are turned of
    if not table_object.configs['use_title_row']:
        if cc1[0] != 0:
            log.debug("METHOD. Title row removed, cc1 was shifted from {} to {}".format(cc1, (0, cc1[1])))
            cc1 = (0, cc1[1])
            table_object.history._title_row_removed = True
    else:
        table_object.history._title_row_removed = False

    # provision for using only the first column of the table as row header
    if table_object.configs['row_header'] is not None:
        row_header = table_object.configs['row_header']
        assert isinstance(row_header, int)
        if table_object.history.prefixed_rows:
            row_header += 1
        left = min(cc1[1], row_header)
        cc1 = (cc1[0], left)
        cc2 = (cc2[0], row_header)

    # provision for using only the first row of the table as column header
    if table_object.configs['col_header'] is not None:
        col_header = table_object.configs['col_header']
        assert isinstance(col_header, int)
        if table_object.history.prefixing_performed and not table_object.history.prefixed_rows:
            col_header += 1
        top = min(cc1[0], col_header)
        cc1 = (top, cc1[1])
        cc2 = (col_header, cc2[1])

    return cc1, cc2


def find_cc3(table_object, cc2):
    """
    Searches for critical cell `CC3`, as the leftmost cell of the first filled row of the data region.

    .. rubric:: Comment on implementation

    There are two options on how to implement the search for `CC3`:

        1. With the possibility of `Notes` rows directly below the header (default):
            * the first half filled row below the header is considered as the start of the data region, just like for the `CC4` cell
            * implemented by Embley et. al.
        2. Without the possibility of `Notes` rows directly below the header:
            * the first row below the header is considered as the start of the data region
            * for scientific tables it might be more common that the first data row only has a single entry
            * this can be chosen my commenting/uncommenting the code within this function

    :param table_object: Input Table object
    :type table_object: ~tabledataextractor.table.table.Table
    :param cc2: Tuple, position of `CC2` cell found with find_cc1_cc2()
    :type cc2: (int,int)
    :return: cc3
    """

    # OPTION 1
    # searching from the top of table for first half-full row, starting with first row below the header:
    n_rows = len(table_object.pre_cleaned_table[cc2[0] + 1:])
    log.debug("n_rows= {}".format(n_rows))
    for row_index in range(cc2[0] + 1, cc2[0] + 1 + n_rows, 1):
        n_full = 0
        n_columns = len(table_object.pre_cleaned_table[row_index, cc2[1] + 1:])
        log.debug("n_columns= {}".format(n_columns))
        for column_index in range(cc2[1] + 1, cc2[1] + 1 + n_columns, 1):
            empty = table_object.pre_cleaned_table_empty[row_index, column_index]
            if not empty:
                n_full += 1
            if n_full >= int(n_columns / 2):
                return row_index, cc2[1] + 1
    raise MIPSError("No CC3 critical cell found! No data region defined.")
    # OPTION 2
    # return (cc2[0]+1,cc2[1]+1)


def find_title_row(table_object):
    """
    Searches for the topmost non-empty row.

    :param table_object: Input Table object
    :type table_object: ~tabledataextractor.table.table.Table
    :return: int
    """
    for row_index, empty_row in enumerate(table_object.pre_cleaned_table_empty):
        if not empty_row.all():
            return row_index


def find_note_cells(table_object, labels_table):
    """
    Searches for all non-empty cells that have not been labelled differently.

    :param table_object: Input Table object
    :type table_object: ~tabledataextractor.table.table.Table
    :param labels_table: table that holds all the labels
    :type labels_table: Numpy array
    :return: Tuple
    """
    for row_index, row in enumerate(labels_table):
        for column_index, cell in enumerate(row):
            if cell == '/' and not table_object.pre_cleaned_table_empty[row_index, column_index]:
                yield row_index, column_index


def prefix_duplicate_labels(table_object, array):
    """
    Prefixes duplicate labels in first row or column where this is possible,
    by adding a new row/column containing the preceding (to the left or above) unique labels, if available.

    Nested prefixing is not supported.

    The algorithm is not completely selective and there might be cases where it's application is undesirable.
    However, on standard datasets it significantly improves table-region classification.

    Algorithm for column headers:

    1. Run MIPS, to find the old header region, without prefixing.
    2. For row in table, can *meaningful* prefixing in this row been done?
        * yes --> do prefixing and go to 3, prefixing of only one row is possible; accept prefixing only if prefixed rows/cells are above the end of the header (not in the data region), the prefixed cells can still be above the header
        * no  --> go to 2, next row
    3. run MIPS to get the new header region
    4. accept prefixing only if the prefixing has not made the header region start lower than before and if it hasn't made the header region wider than before

    The algorithm has been modified from Embley et al., *DOI: 10.1007/s10032-016-0259-1*.

    :param table_object: Input Table object
    :type table_object: ~tabledataextractor.table.table.Table
    :param array: Table to use as input and to do the prefixing on
    :type array: Numpy array
    :return: Table with added rows/columns with prefixes, or, input table, if no prefixing was done

    """

    def unique(data, row_or_column):
        """
        Returns True if data is unique in the given row/column or False if not unique or not present.

        :param data:
        :param row_or_column:
        :return:
        """
        count = 0
        for cell in row_or_column:
            if cell == data:
                count += 1
        if count == 1:
            return True
        else:
            return False

    def prefixed_row_or_column(table):
        """
        Main algorithm for creating prefixed column/row headers.
        If cell is not unique, it is prefixed with the first unique above (for row header) or to the left
        (for column header).

        Returns the row/column containing the prefixes and the position of the row/column where the new row/column
        has to be inserted into the original table.

        This function is getting ugly and could be rewritten with the use of a nice list of tuples,
        for every row/column in the table, we would have a list of distinct elements with their positions in the row/column

        :param table: input table (will not be changed)
        :return: row_index: where the row/column has to be inserted, new_row: the list of prefixes
        """

        unique_prefix = False
        prefixed = False
        row_index = 0
        new_row = []
        for row_index, row in enumerate(table):
            duplicated_row = []
            new_row = []
            for cell_index, cell in enumerate(row):
                # append if unique or empty cell
                if unique(cell, row) or empty_string(cell):
                    duplicated_row.append(cell)
                    new_row.append("")
                else:
                    # find the first unique cell to the left
                    # don't use the first column and first row
                    # as these will presumably be in the stub header region
                    for prefix in reversed(duplicated_row[1:]):
                        # use the prefix if it is unique and not empty
                        if unique(prefix, row) and not empty_string(prefix):
                            unique_prefix = prefix
                            break
                    # prefix the cell and append it to new row
                    if unique_prefix:
                        duplicated_row.append(unique_prefix + "/" + cell)
                        new_row.append(unique_prefix)
                        prefixed = True
                    # else, if no unique prefix was found, just append the original cell,
                    else:
                        duplicated_row.append(cell)
                        new_row.append("")
            # and continue to the next row (if no prefixing has been performed)
            if prefixed:
                break
        if prefixed:
            return row_index, new_row
        else:
            return None

    # MAIN ALGORITHM
    # 1. first, check the MIPS, to see what header we would have gotten without the prefixing
    # note, cc4 couldn't have changed
    log.debug("Prefixing. Attempt to run main MIPS algorithm.")
    try:
        cc1, cc2 = find_cc1_cc2(table_object, find_cc4(table_object), array)
    except (MIPSError, TypeError):
        log.error("Prefixing was not performed due to failure of MIPS algorithm.")
        return array

    # this flag is used for the return value, if it doesn't change the original table is returned
    prefixed = False

    # 2. DO THE PREFIXING
    # prefixing of column headers
    if prefixed_row_or_column(array):
        row_index, new_row = prefixed_row_or_column(array)
        # only perform prefixing if not below of header region (above is allowed!)
        # to allow prefixing even below the old header region cannot be right
        if row_index <= cc2[0]:
            log.debug("Column header prefixing, row_index= {}".format(row_index))
            log.debug("Prefixed row= {}".format(new_row))
            # Prefixing by adding new row:
            prefixed = True
            prefixed_table = np.insert(array, row_index, new_row, axis=0)

    # prefixing of row headers
    if prefixed_row_or_column(array.T):
        column_index, new_column = prefixed_row_or_column(array.T)
        # only perform prefixing if not to the right of header region (to the left is allowed!)
        # to allow prefixing even below the old header region cannot be right
        if column_index <= cc2[1]:
            log.debug("Row header prefixing, column_index= {}".format(column_index))
            log.debug("Prefixed column= {}".format(new_column))
            # Prefixing by adding a new column:
            prefixed = True
            prefixed_table = np.insert(array, column_index, new_column, axis=1)

    # 3. check the headers again, after prefixing
    # note, cc4 couldn't have changed
    if prefixed:
        # if new headers fail, the prefixing has destroyed the table, which is not a HIT table anymore
        try:
            cc1_new, cc2_new = find_cc1_cc2(table_object, find_cc4(table_object), prefixed_table)
        except (MIPSError, TypeError):
            log.debug("Prefixing was not performed because it destroyed the table")
            return array
        # return prefixed_table only if the prefixing has not made the header to start lower,
        # it can end lower (and this is desired and what we want - not to include the data region into the header),
        # but it cannot start lower, because that would mean that we have removed some of the hierarchy and added
        # hierarchy from the left/above into a column/row
        if cc1_new[0] <= cc1[0] and cc1_new[1] <= cc1[1]:
            # Another condition, the header has to end lower than before, not to include at east one
            # lower row/column that was included before
            if cc2_new[0] <= cc2[0] and cc2_new[1] <= cc2[1]:
                table_object.history._prefixing_performed = True
                log.debug("METHOD. Prefixing was performed.")
                if len(prefixed_table.T) > len(array.T):
                    table_object.history._prefixed_rows = True
                return prefixed_table
            else:
                return array
        else:
            return array
    else:
        return array


def duplicate_spanning_cells(table_object, array):
    """
    Duplicates cell contents into appropriate spanning cells. This is sometimes necessary for `.csv` files where
    information has been lost, or, if the source table is not properly formatted.

    Cells outside the row/column header (such as data cells) will not be duplicated.
    MIPS is run to perform a check for that.

    Algorithm according to Nagy and Seth, 2016, in Procs. ICPR 2016, Cancun, Mexico.

    :param table_object: Input Table object
    :type table_object: ~tabledataextractor.table.table.Table
    :param array: Table to use as input
    :type array: Numpy array
    :return: Array with spanning cells copied, if necessary. Alternatively, returns the original table.
    """

    def empty_row(arrayy):
        """Returns 'True' if the whole row is truly empty"""
        for element in arrayy:
            if element:
                return False
        return True

    # running MIPS to find the data region
    log.debug("Spanning cells. Attempt to run MIPS algorithm, to find potential title row.")
    try:
        cc1, cc2 = find_cc1_cc2(table_object, find_cc4(table_object), table_object.pre_cleaned_table)
    except (MIPSError, TypeError):
        log.error("Spanning cells update was not performed due to failure of MIPS algorithm.")
        return array

    log.debug("Spanning cells. Attempt to run main spanning cell algorithm.")
    temp = array.copy()
    top_fill = None
    left_fill = None
    for c in range(0, len(temp.T)):
        flag = 0
        for r in range(cc1[0], len(temp)):
            if temp[r, c]:
                top_fill = temp[r, c]
                flag = 1
            elif flag == 1:
                temp[r, c] = top_fill
            if len(temp) - 1 > r and empty_row(temp[r + 1]):
                flag = 0
    for r in range(cc1[0], len(temp)):
        flag = 0
        for c in range(len(temp.T)):
            if temp[r, c]:
                if (len(temp) - 1 > r and temp[r + 1, c] != temp[r, c]) or temp[r - 1, c] != temp[r, c]:
                    left_fill = temp[r, c]
                    flag = 1
                else:
                    flag = 0
            elif flag == 1:
                temp[r, c] = left_fill
            if len(temp.T) - 1 > c and empty_row(temp.T[c + 1]):
                flag = 0

    # Finding the header regions to make sure the spanning cells additions are not applied in the data region
    # Then, the main MIPS algorithm has to be run
    temp2 = np.copy(temp)
    diff_row_length = 0
    diff_col_length = 0
    if table_object.configs['use_prefixing']:
        temp2 = prefix_duplicate_labels(table_object, temp)
        # reset the prefixing flag
        table_object.history._prefixing_performed = False
        table_object.history._prefixed_rows = False
        diff_row_length = len(temp2) - len(temp)
        diff_col_length = len(temp2.T) - len(temp.T)
    log.debug("Spanning cells. Attempt to run main MIPS algorithm.")
    # disable title row temporarily
    old_title_row_setting = table_object.configs['use_title_row']
    table_object.configs['use_title_row'] = False
    try:
        cc1, cc2 = find_cc1_cc2(table_object, find_cc4(table_object), temp2)
    except (MIPSError, TypeError):
        log.error("Spanning cells update was not performed due to failure of MIPS algorithm.")
        return array
    finally:
        table_object.configs['use_title_row'] = old_title_row_setting

    updated = array.copy()
    # update the original table with values from the updated table if the cells are in the header regions
    # update column header
    for col_header_index in range(cc1[0], cc2[0] + 1 - diff_row_length):
        updated[col_header_index, :] = temp[col_header_index, :]

    # update row header
    for row_header_index in range(cc1[1], cc2[1] + 1 - diff_col_length):
        updated[:, row_header_index] = temp[:, row_header_index]

    # log
    if not np.array_equal(updated, array):
        table_object.history._spanning_cells_extended = True
        log.debug("METHOD. Spanning cells extended.")

    return updated


def header_extension_up(table_object, cc1):
    """
    Extends the header after main MIPS run.

    Algorithm according to Nagy and Seth, 2016, *"Table Headers: An entrance to the data mine"*,
    in Procs. ICPR 2016, Cancun, Mexico.

    :param table_object: Input Table object
    :type table_object: ~tabledataextractor.table.table.Table
    :param cc1: `CC1` critical cell
    :return: cc1_new
    """

    cc1_new_row = None
    cc1_new_col = None

    # add row above the identified column header if it does not consist of cells with identical values and if it
    # adds at least one non-blank cell that has a value different from the cell immediately below it
    current_row = table_object.pre_cleaned_table[cc1[0], :]
    for row_index in range(cc1[0]-1, -1, -1):
        # start after the first column to allow for a title
        if len(np.unique(table_object.pre_cleaned_table[row_index, 1:])) == 1:
            cc1_new_row = row_index+1
        else:
            for col_index, cell in enumerate(table_object.pre_cleaned_table[row_index, :]):
                # remove the first row from this check to preserve a title,
                # if the title is the only non-empty element of the row
                if col_index != 0 and \
                        cell != current_row[col_index] and \
                        not table_object.pre_cleaned_table_empty[row_index, col_index]:
                    current_row = table_object.pre_cleaned_table[row_index, :]
                    cc1_new_row = row_index
                    break
    if cc1_new_row is None:
        cc1_new_row = cc1[0]

    # now do the same for the row headers
    current_col = table_object.pre_cleaned_table[:, cc1[1]]
    for col_index in range(cc1[1]-1, -1, -1):
        if len(np.unique(table_object.pre_cleaned_table[:, col_index])) == 1:
            cc1_new_col = col_index+1
        else:
            for row_index, cell in enumerate(table_object.pre_cleaned_table[:, col_index]):
                if cell != current_col[row_index] and not table_object.pre_cleaned_table_empty[row_index, col_index]:
                    current_col = table_object.pre_cleaned_table[:, col_index]
                    cc1_new_col = col_index
                    break
    if cc1_new_col is None:
        cc1_new_col = cc1[1]

    cc1_new = (cc1_new_row, cc1_new_col)

    # log
    if not cc1_new == cc1:
        table_object.history._header_extended_up = True
        log.debug("METHOD. Header extended upwards.")

    return cc1_new


def header_extension_down(table_object, cc1, cc2, cc4):
    """
    Extends the header downwards, if no prefixing was done and if the appropriate stub header is empty.
    For column-header expansion downwards, only the first cell of the stub header has to be empty.
    For row-header expansion to the right, the whole stub header column above has to be empty.

    :param table_object: Input Table object
    :type table_object: ~tabledataextractor.table.table.Table
    :param cc2: Critical cell `CC2`
    :type cc2: (int, int)
    :param cc1: Critical cell `CC1`
    :type cc1: (int, int)
    :param cc4: Critical cell `CC4`
    :type cc4: (int, int)
    :return: New `cc2`
    """
    cc2_new = cc2
    extended = False

    # only do downwards header extension if no prefixing was done
    if not table_object.history.prefixing_performed:

        # extend column header downwards, changes cc2 row
        # only the first cell of the stub header has to be empty to accept the move downwards
        row_index = cc2[0]
        while row_index <= cc4[0] and empty_string(table_object.pre_cleaned_table[row_index, cc1[1]]):
            row_index += 1
            cc2_new = (row_index - 1, cc2_new[1])
            if cc2_new != cc2:
                extended = True

        if extended:
            table_object.history._header_extended_down = True

        # Check if row header can be shortened now, check duplicate rows accordingly, changes cc2 col
        if extended:
            cc2_new_col = cc2_new[1]
            i = len(table_object.row_header.T)
            while not duplicate_rows(table_object.row_header[:, :i]) and i > 1:
                i -= 1
                if not duplicate_rows(table_object.row_header[:, :i]):
                    cc2_new_col -= 1
            cc2_new = (cc2_new[0], cc2_new_col)
            extended = False

        # extend row header to the right, changes cc2 col
        # this check is more rigorous than above, and all the cells in the stub header have to be empty
        col_index = cc2_new[1]
        while col_index <= cc4[1] and empty_cells(table_object.pre_cleaned_table[cc1[0]:cc2[0]+1, col_index]).all():
            col_index += 1
            if col_index - 1 != cc2_new[1]:
                extended = True
            cc2_new = (cc2_new[0], col_index - 1)

        if extended:
            # Check if column header can be shortened now, changes cc2 row
            cc2_new_row = cc2_new[0]
            i = len(table_object.col_header)
            while not duplicate_columns(table_object.col_header[:i, :]) and i > 1:
                i -= 1
                if not duplicate_columns(table_object.col_header[:i, :]):
                    cc2_new_row -= 1
            cc2_new = (cc2_new_row, cc2_new[1])

        if extended:
            table_object.history._header_extended_down = True

    return cc2_new


def categorize_header(header):
    """
    Performs header categorization (calls the `SymPy` `fact` function) for a given table.

    :param header: header region, Numpy array
    :return: factor_list
    """

    # empty expression and part of the expression that will be factorized
    # these are SymPy expressions
    expression = 0
    part = 0
    for row_index, row in enumerate(header):
        for column_index, cell in enumerate(row):
            if column_index == 0:
                part = Symbol(cell)
            else:
                part = part * Symbol(cell)
        expression = expression + part
    # factorization
    # f = factor(expression, deep=True)
    f = factor_list(expression)
    log.debug("Factorization, initial header: {}".format(expression))
    log.debug("Factorization, factorized header: {}".format(f))
    return f


def build_category_table(table, cc1, cc2, cc3, cc4):
    """
    Build category table for given input table.
    Original header factorization, according to Embley et al., *DOI: 10.1007/s10032-016-0259-1*.
    This version is not used, instead :class:`~tabledataextractor.output.to_pandas.build_category_table` is being used.

    :param table: Table on which to perform the categorization
    :type table: Numpy array
    :param cc1: key MIPS cell
    :param cc2: key MIPS cell
    :param cc3: key MIPS cell
    :param cc4: key MIPS cell
    :return: category table as numpy array
    """

    column_header = table[cc1[0]:cc2[0] + 1, cc3[1]:cc4[1] + 1]
    row_header = table[cc3[0]:cc4[0] + 1, cc1[1]:cc2[1] + 1]
    column_factors = categorize_header(column_header.T)
    row_factors = categorize_header(row_header)


def split_table(table_object):
    """
    Splits table into subtables. Yields :class:`~tabledataextractor.table.table.Table` objects.

    Algorithm:
        If the stub header is repeated in the column header section the table is split up before
        the repeated element.

    :param table_object: Input Table object
    :type table_object: ~tabledataextractor.table.table.Table
    """

    # first, the column header
    i = 0
    # the last row of the column/stub header is not used, as it will be determined as
    # data region by the main MIPS algorithm
    for col_index, column in enumerate(table_object.col_header[:-1].T):
        # the first match is backwards and forwards looking
        if i == 0 and column.size > 0 and \
                table_object.stub_header[:-1].T[0].size > 0 and \
                np.array_equal(column, table_object.stub_header[:-1].T[0]):
            yield table_object._pre_cleaned_table[:, 0:col_index + 1].tolist()
            i += 1
        # every other match is only forwards looking
        if i > 0 and column.size > 0 and \
                table_object.stub_header[:-1].T[0].size > 0 and \
                np.array_equal(column, table_object.stub_header[:-1].T[0]):
            yield table_object._pre_cleaned_table[:, col_index + 1:col_index + i * col_index + 2].tolist()
            i += 1

    # now the same thing for the row header
    i = 0
    for row_index, row in enumerate(table_object.row_header[:, :-1]):
        # the first match is backwards and forwards looking
        if i == 0 and row.size > 0 and \
                table_object.stub_header[0, :-1].size > 0 and \
                np.array_equal(row, table_object.stub_header[0, :-1]):
            yield table_object._pre_cleaned_table[0:row_index + 1, :].tolist()
            i += 1
        # every other match is only forwards looking
        if i > 0 and row.size > 0 and \
                table_object.stub_header[0, :-1].size > 0 \
                and np.array_equal(row, table_object.stub_header[0, :-1]):
            yield table_object._pre_cleaned_table[row_index + 1:row_index + i * row_index + 2, :].tolist()
            i += 1


def find_row_header_table(category_table, stub_header):
    """
    Constructs a Table from the row categories of the original table.

    :param category_table: ~tabledataextractor.table.table.Table.category_table
    :type category_table: list
    :param stub_header: ~tabledataextractor.table.table.Table.stub_header
    :type stub_header: numpy.ndarray
    :return: list
    """
    stub_header = stub_header.tolist()
    raw_table = list()
    for line in stub_header:
        new_line = list()
        for item in line:
            new_line.append(item)
        raw_table.append(new_line)
    for line in category_table:
        new_line = list()
        for item in line[1]:
            new_line.append(item)
        raw_table.append(new_line)
    return raw_table


def clean_row_header(pre_cleaned_table, cc2):
    """
    Cleans the row header by removing duplicate rows that span the whole table.
    """
    unmodified_part = pre_cleaned_table[:cc2[0]+1, :]
    modified_part = pre_cleaned_table[cc2[0]+1:, :]

    # delete duplicate rows that extend over the whole table
    _, indices = np.unique(modified_part, axis=0, return_index=True)
    # for logging only, which rows have been removed
    removed_rows = []
    for row_index in range(0, len(modified_part)):
        if row_index not in indices:
            removed_rows.append(row_index)
    # deletion
    modified_part = modified_part[np.sort(indices)]

    return np.vstack((unmodified_part, modified_part))



# -*- coding: utf-8 -*-
"""
TableDataExtractor
Extracts and standardizes data from tables.
Algorithm from David W. Embley et. Al., DOI: 10.1007/s10032-016-0259-1
jm2111@cam.ac.uk
~~~~~~~~~~~~~~~~~
"""

import logging

__title__ = 'TableDataExtractor'
__version__ = '1.5.11'
__author__ = 'Juraj Mavračić'
__email__ = 'jm2111@cam.ac.uk'
__license__ = 'MIT License'
__copyright__ = 'Copyright 2019 Juraj Mavracic'

# global logging set-up, change the logging level here for development
logging.basicConfig(level=logging.INFO, format='%(levelname)-10s in %(filename)-20s--> %(message)s', handlers=[logging.FileHandler("tde_log.txt", mode='w')])

from tabledataextractor.table.table import Table, TrivialTable




# -*- coding: utf-8 -*-
"""
Indicates to the user which methods have been used on the table.
This should be checked for testing on a sample dataset, to justify the choice of settings for the given domain.

.. codeauthor:: Juraj Mavračić <jm2111@cam.ac.uk>

"""

import logging

log = logging.getLogger(__name__)


class History:
    """
    Stores `True`/`False` for each property, indicating if a method has been used on the particular :class:`~tabledataextractor.table.table.Table` instance.
    """

    def __init__(self):
        self._title_row_removed = False
        self._prefixing_performed = False
        self._prefixed_rows = False
        self._footnotes_copied = False
        self._spanning_cells_extended = False
        self._header_extended_up = False
        self._header_extended_down = False
        self._table_transposed = False
        log.debug("History() object created.")

    @property
    def title_row_removed(self):
        """Indicates whether a title row has been removed from the table."""
        return self._title_row_removed

    @property
    def prefixing_performed(self):
        """Indicates whether prefixing has been performed on the table."""
        return self._prefixing_performed

    @property
    def prefixed_rows(self):
        """Indicates whether prefixing has been performed on the rows (left side)."""
        return self._prefixed_rows

    @property
    def footnotes_copied(self):
        """Indicates whether footnotes have been copied into the table cells."""
        return self._footnotes_copied

    @property
    def spanning_cells_extended(self):
        """
        Indicates whether the content of cells has been duplicated into neighbouring cells,
        in case of cells that are merged cells (spanning cells).
        """
        return self._spanning_cells_extended

    @property
    def header_extended_up(self):
        """Indicates whether the header has been extended upwards, beyond the result obtained by the MIPS
         (*Minimum Indexing Point Search*) algorithm."""
        return self._header_extended_up

    @property
    def header_extended_down(self):
        """Indicates whether the header has been extended downwards, beyond the result obtained by the MIPS
        (*Minimum Indexing Point Search*) algorithm."""
        return self._header_extended_down

    @property
    def table_transposed(self):
        """Indicates whether the table has been transposed."""
        return self._table_transposed

    def __repr__(self):
        out = str()
        out += "title_row_removed       = {}".format(self.title_row_removed)
        out += "\n" + "prefixing_performed     = {}".format(self.prefixing_performed)
        out += "\n" + "prefixed_rows           = {}".format(self.prefixed_rows)
        out += "\n" + "footnotes_copied        = {}".format(self.footnotes_copied)
        out += "\n" + "spanning_cells_extended = {}".format(self.spanning_cells_extended)
        out += "\n" + "header_extended_up      = {}".format(self.header_extended_up)
        out += "\n" + "header_extended_down    = {}".format(self.header_extended_down)
        out += "\n" + "table_transposed        = {}".format(self._table_transposed)
        return out










# -*- coding: utf-8 -*-
"""
Tests full tables - default config - only correct outcomes

.. codeauthor:: Juraj Mavračić (jm2111@cam.ac.uk)

"""

import unittest
import logging
from tabledataextractor import Table

log = logging.getLogger(__name__)


class TestTDE(unittest.TestCase):

    def do_table(self, input_path, expected_labels, expected_category_table):
        log.debug("Test TDE, Table: {}".format(input_path))
        table = Table(input_path)
        table.print()
        labels = table.labels.tolist()
        category_table = table.category_table
        self.assertListEqual(expected_labels, labels)
        self.assertListEqual(expected_category_table, category_table)

    def do_subtable(self, input_path, subtable_number, expected_labels, expected_category_table):
        log.debug("Test TDE, SubTable: {}".format(input_path))
        table = Table(input_path)
        table.subtables[subtable_number].print_raw_table()
        labels = table.subtables[subtable_number].labels.tolist()
        category_table = table.subtables[subtable_number].category_table
        self.assertListEqual(expected_labels, labels)
        self.assertListEqual(expected_category_table, category_table)

    def test_01(self):
        input_path = './tests/data/table_example1.csv'
        expected_labels = [['StubHeader', 'StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'], ['StubHeader', 'StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'], ['RowHeader', 'RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data']]
        expected_category_table = [['4.64', ['Computational', 'This study'], ['Rutile', 'a = b (A)']], ['2.99', ['Computational', 'This study'], ['Rutile', 'c (A)']], ['0.305', ['Computational', 'This study'], ['Rutile', 'u']], ['3.83', ['Computational', 'This study'], ['Anatase', 'a = b (A)']], ['9.62', ['Computational', 'This study'], ['Anatase', 'c (A)']], ['0.208', ['Computational', 'This study'], ['Anatase', 'u']], ['4.67', ['Computational', 'GGA [25]'], ['Rutile', 'a = b (A)']], ['2.97', ['Computational', 'GGA [25]'], ['Rutile', 'c (A)']], ['0.305', ['Computational', 'GGA [25]'], ['Rutile', 'u']], ['3.80', ['Computational', 'GGA [25]'], ['Anatase', 'a = b (A)']], ['9.67', ['Computational', 'GGA [25]'], ['Anatase', 'c (A)']], ['0.207', ['Computational', 'GGA [25]'], ['Anatase', 'u']], ['4.63', ['Computational', 'GGA [26]'], ['Rutile', 'a = b (A)']], ['2.98', ['Computational', 'GGA [26]'], ['Rutile', 'c (A)']], ['0.305', ['Computational', 'GGA [26]'], ['Rutile', 'u']], ['NoValue', ['Computational', 'GGA [26]'], ['Anatase', 'a = b (A)']], ['NoValue', ['Computational', 'GGA [26]'], ['Anatase', 'c (A)']], ['NoValue', ['Computational', 'GGA [26]'], ['Anatase', 'u']], ['NoValue', ['Computational', 'HF [27]'], ['Rutile', 'a = b (A)']], ['NoValue', ['Computational', 'HF [27]'], ['Rutile', 'c (A)']], ['NoValue', ['Computational', 'HF [27]'], ['Rutile', 'u']], ['3.76', ['Computational', 'HF [27]'], ['Anatase', 'a = b (A)']], ['9.85', ['Computational', 'HF [27]'], ['Anatase', 'c (A)']], ['0.202', ['Computational', 'HF [27]'], ['Anatase', 'u']], ['4.594', ['Experimental', 'Expt. [23]'], ['Rutile', 'a = b (A)']], ['2.958', ['Experimental', 'Expt. [23]'], ['Rutile', 'c (A)']], ['0.305', ['Experimental', 'Expt. [23]'], ['Rutile', 'u']], ['3.785', ['Experimental', 'Expt. [23]'], ['Anatase', 'a = b (A)']], ['9.514', ['Experimental', 'Expt. [23]'], ['Anatase', 'c (A)']], ['0.207', ['Experimental', 'Expt. [23]'], ['Anatase', 'u']]]
        self.do_table(input_path, expected_labels, expected_category_table)

    def test_02(self):
        input_path = './tests/data/table_example10.csv'
        expected_labels = [['TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle'], ['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'], ['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data']]
        expected_category_table = [['1647218', ['2003'], ['Short messages/thousands', 'Short messages/thousands']], ['24.3', ['2003'], ['Change %', 'A']], ['347', ['2003'], [' Short messages/subscription', ' Short messages/subscription']], ['2314', ['2003'], ['Multimedia messages/thousands', 'Multimedia messages/thousands']], ['NoValue', ['2003'], ['Change %', 'B']], ['2193498', ['2004'], ['Short messages/thousands', 'Short messages/thousands']], ['33.2', ['2004'], ['Change %', 'A']], ['439', ['2004'], [' Short messages/subscription', ' Short messages/subscription']], ['7386', ['2004'], ['Multimedia messages/thousands', 'Multimedia messages/thousands']], ['219.2', ['2004'], ['Change %', 'B']]]
        self.do_table(input_path, expected_labels, expected_category_table)

    def test_03(self):
        input_path = './tests/data/table_example13.csv'
        expected_labels = [['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data']]
        expected_category_table = [['1.82 ± 0.3', ['NAsAs'], ['As40Te60']],
                                   ['1.07 ± 0.3', ['NAsAs'], ['As30Cu10Te60']],
                                   ['1.13 ± 0.3', ['NAsAs'], ['As20Cu20Te60']],
                                   ['1.18 ± 0.3', ['NAsAs'], ['As15Cu25Te60']],
                                   ['0.79 ± 0.3', ['NAsAs'], ['As10Cu30Te60']],
                                   ['NoValue', ['NAsCu'], ['As40Te60']],
                                   ['0.37 ± 0.2', ['NAsCu'], ['As30Cu10Te60']],
                                   ['0.71 ± 0.25', ['NAsCu'], ['As20Cu20Te60']],
                                   ['0.85 ± 0.25', ['NAsCu'], ['As15Cu25Te60']],
                                   ['0.51 ± 0.3', ['NAsCu'], ['As10Cu30Te60']],
                                   ['1.34 ± 0.3', ['NAsTe'], ['As40Te60']],
                                   ['1.58 ± 0.3', ['NAsTe'], ['As30Cu10Te60']],
                                   ['1.80 ± 0.3', ['NAsTe'], ['As20Cu20Te60']],
                                   ['1.94 ± 0.4', ['NAsTe'], ['As15Cu25Te60']],
                                   ['2.76 ± 0.6', ['NAsTe'], ['As10Cu30Te60']],
                                   ['NoValue', ['NCuAs'], ['As40Te60']],
                                   ['1.11 ± 0.6', ['NCuAs'], ['As30Cu10Te60']],
                                   ['0.71 ± 0.25', ['NCuAs'], ['As20Cu20Te60']],
                                   ['0.51 ± 0.15', ['NCuAs'], ['As15Cu25Te60']],
                                   ['0.17 ± 0.1', ['NCuAs'], ['As10Cu30Te60']],
                                   ['NoValue', ['NCuCu'], ['As40Te60']],
                                   ['1.07 ± 0.6', ['NCuCu'], ['As30Cu10Te60']],
                                   ['1.69 ± 0.5', ['NCuCu'], ['As20Cu20Te60']],
                                   ['1.83 ± 0.5', ['NCuCu'], ['As15Cu25Te60']],
                                   ['3.11 ± 0.7', ['NCuCu'], ['As10Cu30Te60']],
                                   ['NoValue', ['NCuTe'], ['As40Te60']],
                                   ['2.32 ± 0.6', ['NCuTe'], ['As30Cu10Te60']],
                                   ['2.01 ± 0.6', ['NCuTe'], ['As20Cu20Te60']],
                                   ['2.46 ± 0.6', ['NCuTe'], ['As15Cu25Te60']],
                                   ['2.17 ± 0.5', ['NCuTe'], ['As10Cu30Te60']],
                                   ['0.89', ['NTeAs'], ['As40Te60']],
                                   ['0.79 ± 0.25', ['NTeAs'], ['As30Cu10Te60']],
                                   ['0.60 ± 0.2', ['NTeAs'], ['As20Cu20Te60']],
                                   ['0.49 ± 0.2', ['NTeAs'], ['As15Cu25Te60']],
                                   ['0.46 ± 0.1', ['NTeAs'], ['As10Cu30Te60']],
                                   ['NoValue', ['NTeCu'], ['As40Te60']],
                                   ['0.39 ± 0.2', ['NTeCu'], ['As30Cu10Te60']],
                                   ['0.67 ± 0.2', ['NTeCu'], ['As20Cu20Te60']],
                                   ['1.03 ± 0.25', ['NTeCu'], ['As15Cu25Te60']],
                                   ['1.08 ± 0.25', ['NTeCu'], ['As10Cu30Te60']],
                                   ['1.01', ['NTeTe'], ['As40Te60']],
                                   ['1.14 ± 0.2', ['NTeTe'], ['As30Cu10Te60']],
                                   ['1.43 ± 0.2', ['NTeTe'], ['As20Cu20Te60']],
                                   ['1.44 ± 0.2', ['NTeTe'], ['As15Cu25Te60']],
                                   ['1.65 ± 0.2', ['NTeTe'], ['As10Cu30Te60']],
                                   ['3.16 ± 0.3', ['NAsAs + NAsTe'], ['As40Te60']],
                                   ['2.65 ± 0.3', ['NAsAs + NAsTe'], ['As30Cu10Te60']],
                                   ['2.93 ± 0.3', ['NAsAs + NAsTe'], ['As20Cu20Te60']],
                                   ['3.12 ± 0.3', ['NAsAs + NAsTe'], ['As15Cu25Te60']],
                                   ['3.55 ± 0.6', ['NAsAs + NAsTe'], ['As10Cu30Te60']],
                                   ['1.90 ± 0.2', ['NTeAs + NTeTe'], ['As40Te60']],
                                   ['1.93 ± 0.2', ['NTeAs + NTeTe'], ['As30Cu10Te60']],
                                   ['2.03 ± 0.2', ['NTeAs + NTeTe'], ['As20Cu20Te60']],
                                   ['1.93 ± 0.2', ['NTeAs + NTeTe'], ['As15Cu25Te60']],
                                   ['2.11 ± 0.2', ['NTeAs + NTeTe'], ['As10Cu30Te60']],
                                   ['3.16 ± 0.3', ['NAs'], ['As40Te60']],
                                   ['3.02 ± 0.3', ['NAs'], ['As30Cu10Te60']],
                                   ['3.64 ± 0.3', ['NAs'], ['As20Cu20Te60']],
                                   ['3.97 ± 0.3', ['NAs'], ['As15Cu25Te60']],
                                   ['4.06 ± 0.6', ['NAs'], ['As10Cu30Te60']],
                                   ['NoValue', ['NCu'], ['As40Te60']],
                                   ['4.50 ± 1.0', ['NCu'], ['As30Cu10Te60']],
                                   ['4.41 + 0.5', ['NCu'], ['As20Cu20Te60']],
                                   ['4.80 ± 0.5', ['NCu'], ['As15Cu25Te60']],
                                   ['5.45 ± 0.6', ['NCu'], ['As10Cu30Te60']],
                                   ['1.90 ± 0.2', ['NTe'], ['As40Te60']],
                                   ['2.32 ± 0.2', ['NTe'], ['As30Cu10Te60']],
                                   ['2.70 ± 0.2', ['NTe'], ['As20Cu20Te60']],
                                   ['2.96 ± 0.2', ['NTe'], ['As15Cu25Te60']],
                                   ['3.19 ± 0.3', ['NTe'], ['As10Cu30Te60']]]
        self.do_table(input_path, expected_labels, expected_category_table)

    def test_04(self):
        input_path = './tests/data/table_example3.csv'
        expected_labels = [['TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle'], ['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'], ['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data']]
        expected_category_table = [['3735', ['Norway'], ['Million dollar', '2007']], ['4006', ['Norway'], ['Million dollar', '2008']], ['4081', ['Norway'], ['Million dollar', '2009']], ['4580', ['Norway'], ['Million dollar', '2010*']], ['4936', ['Norway'], ['Million dollar', '2011*']], ['0.95', ['Norway'], ['Percentage of GNI', '2007']], ['0.89', ['Norway'], ['Percentage of GNI', '2008']], ['1.06', ['Norway'], ['Percentage of GNI', '2009']], ['1.1', ['Norway'], ['Percentage of GNI', '2010*']], ['1', ['Norway'], ['Percentage of GNI', '2011*']], ['2562', ['Denmark'], ['Million dollar', '2007']], ['2803', ['Denmark'], ['Million dollar', '2008']], ['2810', ['Denmark'], ['Million dollar', '2009']], ['2871', ['Denmark'], ['Million dollar', '2010*']], ['2981', ['Denmark'], ['Million dollar', '2011*']], ['0.81', ['Denmark'], ['Percentage of GNI', '2007']], ['0.82', ['Denmark'], ['Percentage of GNI', '2008']], ['0.88', ['Denmark'], ['Percentage of GNI', '2009']], ['0.91', ['Denmark'], ['Percentage of GNI', '2010*']], ['0.86', ['Denmark'], ['Percentage of GNI', '2011*']], ['2669', ['Australia'], ['Million dollar', '2007']], ['2954', ['Australia'], ['Million dollar', '2008']], ['2762', ['Australia'], ['Million dollar', '2009']], ['3826', ['Australia'], ['Million dollar', '2010*']], ['4799', ['Australia'], ['Million dollar', '2011*']], ['0.32', ['Australia'], ['Percentage of GNI', '2007']], ['0.32', ['Australia'], ['Percentage of GNI', '2008']], ['0.29', ['Australia'], ['Percentage of GNI', '2009']], ['0.32', ['Australia'], ['Percentage of GNI', '2010*']], ['0.35', ['Australia'], ['Percentage of GNI', '2011*']], ['320', ['New Zealand'], ['Million dollar', '2007']], ['348', ['New Zealand'], ['Million dollar', '2008']], ['309', ['New Zealand'], ['Million dollar', '2009']], ['342', ['New Zealand'], ['Million dollar', '2010*']], ['429', ['New Zealand'], ['Million dollar', '2011*']], ['0.27', ['New Zealand'], ['Percentage of GNI', '2007']], ['0.3', ['New Zealand'], ['Percentage of GNI', '2008']], ['0.28', ['New Zealand'], ['Percentage of GNI', '2009']], ['0.26', ['New Zealand'], ['Percentage of GNI', '2010*']], ['0.28', ['New Zealand'], ['Percentage of GNI', '2011*']], ['104206', ['OECD/DAC1'], ['Million dollar', '2007']], ['121954', ['OECD/DAC1'], ['Million dollar', '2008']], ['119778', ['OECD/DAC1'], ['Million dollar', '2009']], ['128465', ['OECD/DAC1'], ['Million dollar', '2010*']], ['133526', ['OECD/DAC1'], ['Million dollar', '2011*']], ['0.27', ['OECD/DAC1'], ['Percentage of GNI', '2007']], ['0.3', ['OECD/DAC1'], ['Percentage of GNI', '2008']], ['0.31', ['OECD/DAC1'], ['Percentage of GNI', '2009']], ['0.32', ['OECD/DAC1'], ['Percentage of GNI', '2010*']], ['0.31', ['OECD/DAC1'], ['Percentage of GNI', '2011*']]]
        self.do_table(input_path, expected_labels, expected_category_table)

    def test_05(self):
        input_path = './tests/data/table_example4.csv'
        expected_labels = [['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data']]
        expected_category_table = [['1647218', ['2003'], ['Short messages/thousands']], ['24.3', ['2003'], ['A Change %']], ['347', ['2003'], [' Short messages/subscription']], ['2314', ['2003'], ['Multimedia messages/thousands']], ['NoValue', ['2003'], ['B Change %']], ['2193498', ['2004'], ['Short messages/thousands']], ['33.2', ['2004'], ['A Change %']], ['439', ['2004'], [' Short messages/subscription']], ['7386', ['2004'], ['Multimedia messages/thousands']], ['219.2', ['2004'], ['B Change %']]]
        self.do_table(input_path, expected_labels, expected_category_table)

    def test_06(self):
        input_path = './tests/data/table_example6.csv'
        expected_labels = [['StubHeader', 'ColHeader'], ['RowHeader', 'Data'], ['RowHeader', 'Data'], ['RowHeader', 'Data'], ['RowHeader', 'Data'], ['RowHeader', 'Data'], ['RowHeader', 'Data'], ['RowHeader', 'Data'], ['RowHeader', 'Data'], ['RowHeader', 'Data'], ['RowHeader', 'Data'], ['RowHeader', 'Data']]
        expected_category_table = [['6.5 K', ['Bi6Tl3'], ['Tc']], ['5.5', ['Sb2Tl7'], ['Tc']], ['7.2', ['Na2Pb5'], ['Tc']], ['3.8', ['Hg5Tl7'], ['Tc']], ['1.84', ['Au2Bi'], ['Tc']], ['1.6', ['CuS'], ['Tc']], ['1.3', ['VN'], ['Tc']], ['2.8', ['WC'], ['Tc']], ['2.05', ['W2C'], ['Tc']], ['7.7', ['MoC'], ['Tc']], ['2.4', ['Mo2C'], ['Tc']]]
        self.do_subtable(input_path, 0, expected_labels, expected_category_table)
        expected_labels = [['StubHeader', 'ColHeader'], ['RowHeader', 'Data'], ['RowHeader', 'Data'], ['RowHeader', 'Data'], ['RowHeader', 'Data'], ['RowHeader', 'Data'], ['RowHeader', 'Data'], ['RowHeader', 'Data'], ['RowHeader', 'Data'], ['RowHeader', 'Data'], ['RowHeader', 'Data'], ['RowHeader', 'Data']]
        expected_category_table = [['1.4 K', ['TiN'], ['Tc']], ['1.1', ['TiC'], ['Tc']], ['9.2', ['TaC'], ['Tc']], ['10.1', ['NbC'], ['Tc']], ['2.82', ['ZrB'], ['Tc']], ['4.2', ['TaSi'], ['Tc']], ['4.1', ['PbS'], ['Tc']], ['8.4', ['Pb-As alloy'], ['Tc']], ['8.5', ['Pb-Sn-Bi'], ['Tc']], ['9.0', ['Pb-As-Bi'], ['Tc']], ['8.9', ['Pb-Bi-Sb'], ['Tc']]]
        self.do_subtable(input_path, 1, expected_labels, expected_category_table)

    def test_07(self):
        input_path = './tests/data/table_example7.csv'
        expected_labels = [['TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle'], ['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'], ['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader & FNref', 'ColHeader & FNref', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader & FNref', 'ColHeader & FNref'], ['FNref', '/', '/', '/', '/', '/', '/', '/', '/', '/', '/'], ['RowHeader & FNref', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader & FNref', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'], ['FNprefix & FNtext', 'Note', '/', '/', '/', '/', '/', '/', '/', '/', '/'], ['FNprefix', 'FNtext', '/', '/', '/', '/', '/', '/', '/', '/', '/'], ['Note', '/', '/', '/', '/', '/', '/', '/', '/', '/', '/'], ['FNprefix & FNtext', '/', '/', '/', '/', '/', '/', '/', '/', '/', '/']]
        expected_category_table = [['3735', ["Norway DAC-countries are members of OECD's Development Assis"], ['Million dollar', '2007']], ['4081', ["Norway DAC-countries are members of OECD's Development Assis"], ['Million dollar', '2009']], ['4006', ["Norway DAC-countries are members of OECD's Development Assis"], ['Million dollar', '2008']], ['4580', ["Norway DAC-countries are members of OECD's Development Assis"], ['Million dollar', '2010 Source: OECD ']], ['4936', ["Norway DAC-countries are members of OECD's Development Assis"], ['Million dollar', '2011 Source: OECD ']], ['0.95', ["Norway DAC-countries are members of OECD's Development Assis"], ['Percentage of GNI', '2007']], ['0.89', ["Norway DAC-countries are members of OECD's Development Assis"], ['Percentage of GNI', '2008']], ['1.06', ["Norway DAC-countries are members of OECD's Development Assis"], ['Percentage of GNI', '2009']], ['1.1', ["Norway DAC-countries are members of OECD's Development Assis"], ['Percentage of GNI', '2010 Source: OECD ']], ['1', ["Norway DAC-countries are members of OECD's Development Assis"], ['Percentage of GNI', '2011 Source: OECD ']], ['2562', ['Denmark'], ['Million dollar', '2007']], ['2810', ['Denmark'], ['Million dollar', '2009']], ['2803', ['Denmark'], ['Million dollar', '2008']], ['2871', ['Denmark'], ['Million dollar', '2010 Source: OECD ']], ['2981', ['Denmark'], ['Million dollar', '2011 Source: OECD ']], ['0.81', ['Denmark'], ['Percentage of GNI', '2007']], ['0.82', ['Denmark'], ['Percentage of GNI', '2008']], ['0.88', ['Denmark'], ['Percentage of GNI', '2009']], ['0.91', ['Denmark'], ['Percentage of GNI', '2010 Source: OECD ']], ['0.86', ['Denmark'], ['Percentage of GNI', '2011 Source: OECD ']], ['2669', ['Australia'], ['Million dollar', '2007']], ['2762', ['Australia'], ['Million dollar', '2009']], ['2954', ['Australia'], ['Million dollar', '2008']], ['3826', ['Australia'], ['Million dollar', '2010 Source: OECD ']], ['4799', ['Australia'], ['Million dollar', '2011 Source: OECD ']], ['0.32', ['Australia'], ['Percentage of GNI', '2007']], ['0.32', ['Australia'], ['Percentage of GNI', '2008']], ['0.29', ['Australia'], ['Percentage of GNI', '2009']], ['0.32', ['Australia'], ['Percentage of GNI', '2010 Source: OECD ']], ['0.35', ['Australia'], ['Percentage of GNI', '2011 Source: OECD ']], ['320', ['New Zealand'], ['Million dollar', '2007']], ['309', ['New Zealand'], ['Million dollar', '2009']], ['348', ['New Zealand'], ['Million dollar', '2008']], ['342', ['New Zealand'], ['Million dollar', '2010 Source: OECD ']], ['429', ['New Zealand'], ['Million dollar', '2011 Source: OECD ']], ['0.27', ['New Zealand'], ['Percentage of GNI', '2007']], ['0.3', ['New Zealand'], ['Percentage of GNI', '2008']], ['0.28', ['New Zealand'], ['Percentage of GNI', '2009']], ['0.26', ['New Zealand'], ['Percentage of GNI', '2010 Source: OECD ']], ['0.28', ['New Zealand'], ['Percentage of GNI', '2011 Source: OECD ']], ['104206', ['OECD/DAC This is a description'], ['Million dollar', '2007']], ['119778', ['OECD/DAC This is a description'], ['Million dollar', '2009']], ['121954', ['OECD/DAC This is a description'], ['Million dollar', '2008']], ['128465', ['OECD/DAC This is a description'], ['Million dollar', '2010 Source: OECD ']], ['133526', ['OECD/DAC This is a description'], ['Million dollar', '2011 Source: OECD ']], ['0.27', ['OECD/DAC This is a description'], ['Percentage of GNI', '2007']], ['0.3', ['OECD/DAC This is a description'], ['Percentage of GNI', '2008']], ['0.31', ['OECD/DAC This is a description'], ['Percentage of GNI', '2009']], ['0.32', ['OECD/DAC This is a description'], ['Percentage of GNI', '2010 Source: OECD ']], ['0.31', ['OECD/DAC This is a description'], ['Percentage of GNI', '2011 Source: OECD ']]]
        self.do_table(input_path, expected_labels, expected_category_table)

    def test_08(self):
        input_path = './tests/data/table_example8.csv'
        expected_labels = [['TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle'], ['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'], ['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data']]
        expected_category_table = [['1647218', ['2003'], ['', 'Short messages/thousands']],
                                   ['24.3', ['2003'], ['Short messages/thousands', 'Change %']],
                                   ['347', ['2003'], ['', 'Short messages/subscription']],
                                   ['2314', ['2003'], ['', 'Multimedia messages/thousands']],
                                   ['NoValue', ['2003'], ['Multimedia messages/thousands', 'Change %']],
                                   ['2193498', ['2004'], ['', 'Short messages/thousands']],
                                   ['33.2', ['2004'], ['Short messages/thousands', 'Change %']],
                                   ['439', ['2004'], ['', 'Short messages/subscription']],
                                   ['7386', ['2004'], ['', 'Multimedia messages/thousands']],
                                   ['219.2', ['2004'], ['Multimedia messages/thousands', 'Change %']]]
        self.do_table(input_path, expected_labels, expected_category_table)

    def test_09(self):
        input_path = './tests/data/table_example9.csv'
        expected_labels = [['TableTitle', 'TableTitle', 'TableTitle', 'TableTitle'], ['StubHeader', 'StubHeader', 'ColHeader', 'ColHeader'], ['RowHeader', 'RowHeader', 'Data', 'Data'], ['RowHeader', 'RowHeader', 'Data', 'Data'], ['RowHeader', 'RowHeader', 'Data', 'Data'], ['RowHeader', 'RowHeader', 'Data', 'Data'], ['RowHeader', 'RowHeader', 'Data', 'Data']]
        expected_category_table = [['1647218', ['', 'Short messages/thousands'], ['2003']],
                                   ['2193498', ['', 'Short messages/thousands'], ['2004']],
                                   ['24.3', ['Short messages/thousands', 'Change %'], ['2003']],
                                   ['33.2', ['Short messages/thousands', 'Change %'], ['2004']],
                                   ['347', ['', 'Short messages/subscription'], ['2003']],
                                   ['439', ['', 'Short messages/subscription'], ['2004']],
                                   ['2314', ['', 'Multimedia messages/thousands'], ['2003']],
                                   ['7386', ['', 'Multimedia messages/thousands'], ['2004']],
                                   ['NoValue', ['Multimedia messages/thousands', 'Change %'], ['2003']],
                                   ['219.2', ['Multimedia messages/thousands', 'Change %'], ['2004']]]
        self.do_table(input_path, expected_labels, expected_category_table)

    def test_10(self):
        input_path = './tests/data/table_example_footnotes.csv'
        expected_labels = [['TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle'], ['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'], ['StubHeader', 'ColHeader', 'ColHeader & FNref', 'ColHeader & FNref & FNref', 'ColHeader', 'ColHeader & FNref', 'ColHeader & FNref'], ['RowHeader & FNref', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader & FNref', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader & FNref', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'], ['FNprefix', 'FNtext', '/', '/', '/', '/', '/'], ['FNprefix & FNtext', 'Note', 'Note', '/', '/', '/', '/'], ['FNprefix & FNtext', '/', '/', '/', '/', '/', '/'], ['FNprefix', '/', '/', '/', '/', '/', '/'], ['FNprefix & FNtext', '/', '/', '/', '/', '/', '/'], ['Note', 'Note', '/', '/', '/', '/', '/']]
        expected_category_table = [['3735', [' whataboutthis '], ['Million dollar', '2007']], ['4580', [' whataboutthis '], ['Million dollar', '2010 Test ']], ['4936', [' whataboutthis '], ['Million dollar', '2011 Test   whataboutthis ']], ['0.95', [' whataboutthis '], ['Percentage of GNI', '2007']], ['1.1', [' whataboutthis '], ['Percentage of GNI', '2010 Source: OECD. ']], ['1', [' whataboutthis '], ['Percentage of GNI', '2011 Source: OECD. ']], ['2669', ['2'], ['Million dollar', '2007']], ['3826', ['2'], ['Million dollar', '2010 Test ']], ['4799', ['2'], ['Million dollar', '2011 Test   whataboutthis ']], ['0.32', ['2'], ['Percentage of GNI', '2007']], ['0.32', ['2'], ['Percentage of GNI', '2010 Source: OECD. ']], ['0.35', ['2'], ['Percentage of GNI', '2011 Source: OECD. ']], ['320', ['New Zealand '], ['Million dollar', '2007']], ['342', ['New Zealand '], ['Million dollar', '2010 Test ']], ['429', ['New Zealand '], ['Million dollar', '2011 Test   whataboutthis ']], ['0.27', ['New Zealand '], ['Percentage of GNI', '2007']], ['0.26', ['New Zealand '], ['Percentage of GNI', '2010 Source: OECD. ']], ['0.28', ['New Zealand '], ['Percentage of GNI', '2011 Source: OECD. ']], ['104206', ['OECD/DAC Footnote text.'], ['Million dollar', '2007']], ['128465', ['OECD/DAC Footnote text.'], ['Million dollar', '2010 Test ']], ['133526', ['OECD/DAC Footnote text.'], ['Million dollar', '2011 Test   whataboutthis ']], ['0.27', ['OECD/DAC Footnote text.'], ['Percentage of GNI', '2007']], ['0.32', ['OECD/DAC Footnote text.'], ['Percentage of GNI', '2010 Source: OECD. ']], ['0.31', ['OECD/DAC Footnote text.'], ['Percentage of GNI', '2011 Source: OECD. ']]]
        self.do_table(input_path, expected_labels, expected_category_table)

    def test_11(self):
        input_path = './tests/data/te_01.csv'
        expected_labels = [['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data']]
        expected_category_table = [['4.64', ['This study'], ['A']], ['2.99', ['This study'], ['B']], ['0.305', ['This study'], ['C']], ['3.83', ['This study'], ['D']], ['9.62', ['This study'], ['E']], ['0.208', ['This study'], ['F']]]
        self.do_table(input_path, expected_labels, expected_category_table)

    def test_12(self):
        input_path = './tests/data/te_02.csv'
        expected_labels = [['StubHeader', 'ColHeader'], ['RowHeader', 'Data'], ['RowHeader', 'Data'], ['RowHeader', 'Data'], ['RowHeader', 'Data'], ['RowHeader', 'Data'], ['RowHeader', 'Data']]
        expected_category_table = [['4.64', ['A'], ['This study']], ['2.99', ['B'], ['This study']], ['0.305', ['C'], ['This study']], ['3.83', ['D'], ['This study']], ['9.62', ['E'], ['This study']], ['0.208', ['F'], ['This study']]]
        self.do_table(input_path, expected_labels, expected_category_table)

    def test_13(self):
        input_path = './tests/data/te_04.csv'
        expected_labels = [['TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle'], ['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'], ['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'], ['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data']]
        expected_category_table = [['4869', ['1990'], ['School', 'School', 'School']], ['2189', ['1990'], ['Pupils', 'Pre-primary', 'Pre-primary']], ['389410', ['1990'], ['Pupils', 'Grades', '6 Jan']], ['197719', ['1990'], ['Pupils', 'Grades', '9 Jul']], ['NoValue', ['1990'], ['Pupils', 'Additional', 'Additional']], ['592920', ['1990'], ['Pupils', 'Total', 'Total']], ['67427', ['1990'], ['Grade 1', 'Grade 1', 'Grade 1']], ['61054', ['1990'], ['Leaving certificates', 'Leaving certificates', 'Leaving certificates']], ['4861', ['1991'], ['School', 'School', 'School']], ['2181', ['1991'], ['Pupils', 'Pre-primary', 'Pre-primary']], ['389411', ['1991'], ['Pupils', 'Grades', '6 Jan']], ['197711', ['1991'], ['Pupils', 'Grades', '9 Jul']], ['3601', ['1991'], ['Pupils', 'Additional', 'Additional']], ['592921', ['1991'], ['Pupils', 'Total', 'Total']], ['67421', ['1991'], ['Grade 1', 'Grade 1', 'Grade 1']], ['NoValue', ['1991'], ['Leaving certificates', 'Leaving certificates', 'Leaving certificates']]]
        self.do_table(input_path, expected_labels, expected_category_table)

    def test_14(self):
        input_path = './tests/data/te_05.csv'
        expected_labels = [['TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle'], ['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'], ['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data']]
        expected_category_table = [['0', ['Total'], ['', 'Add']], ['597', ['Total'], ['Total', 'justice']], ['264', ['Total'], ['Police', 'protec-']], ['123', ['Total'], ['Judicial', 'and']], ['210', ['Total'], ['', 'Correc-']], ['23', ['Alabama'], ['', 'Add']], ['420', ['Alabama'], ['Total', 'justice']], ['201', ['Alabama'], ['Police', 'protec-']], ['78', ['Alabama'], ['Judicial', 'and']], ['141', ['Alabama'], ['', 'Correc-']], ['0', ['Alaska'], ['', 'Add']], ['894', ['Alaska'], ['Total', 'justice']], ['318', ['Alaska'], ['Police', 'protec-']], ['263', ['Alaska'], ['Judicial', 'and']], ['313', ['Alaska'], ['', 'Correc-']], ['0', ['Arizona'], ['', 'Add']], ['643', ['Arizona'], ['Total', 'justice']], ['275', ['Arizona'], ['Police', 'protec-']], ['137', ['Arizona'], ['Judicial', 'and']], ['232', ['Arizona'], ['', 'Correc-']]]
        self.do_table(input_path, expected_labels, expected_category_table)

    def test_15(self):
        input_path = './tests/data/te_06.csv'
        expected_labels = [['StubHeader', 'ColHeader', 'ColHeader'], ['RowHeader', 'Data', 'Data'], ['RowHeader', 'Data', 'Data'], ['RowHeader', 'Data', 'Data'], ['RowHeader', 'Data', 'Data'], ['RowHeader', 'Data', 'Data'], ['RowHeader', 'Data', 'Data'], ['RowHeader', 'Data', 'Data'], ['RowHeader', 'Data', 'Data'], ['RowHeader', 'Data', 'Data'], ['RowHeader', 'Data', 'Data'], ['RowHeader', 'Data', 'Data']]
        expected_category_table = [['6.5', ['Bi6Tl3'], ['Tc']], ['x', ['Bi6Tl3'], ['A']], ['5.5', ['Sb2Tl7'], ['Tc']], ['y', ['Sb2Tl7'], ['A']], ['7.2', ['Na2Pb5'], ['Tc']], ['z', ['Na2Pb5'], ['A']], ['3.8', ['Hg5Tl7'], ['Tc']], ['x', ['Hg5Tl7'], ['A']], ['1.84', ['Au2Bi'], ['Tc']], ['x', ['Au2Bi'], ['A']], ['1.6', ['CuS'], ['Tc']], ['x', ['CuS'], ['A']], ['1.3', ['VN'], ['Tc']], ['x', ['VN'], ['A']], ['2.8', ['WC'], ['Tc']], ['x', ['WC'], ['A']], ['2.05', ['W2C'], ['Tc']], ['x', ['W2C'], ['A']], ['7.7', ['MoC'], ['Tc']], ['x', ['MoC'], ['A']], ['2.4', ['Mo2C'], ['Tc']], ['x', ['Mo2C'], ['A']]]
        self.do_subtable(input_path, 0, expected_labels, expected_category_table)
        expected_labels = [['StubHeader', 'ColHeader', 'ColHeader'], ['RowHeader', 'Data', 'Data'], ['RowHeader', 'Data', 'Data'], ['RowHeader', 'Data', 'Data'], ['RowHeader', 'Data', 'Data'], ['RowHeader', 'Data', 'Data'], ['RowHeader', 'Data', 'Data'], ['RowHeader', 'Data', 'Data'], ['RowHeader', 'Data', 'Data'], ['RowHeader', 'Data', 'Data'], ['RowHeader', 'Data', 'Data'], ['RowHeader', 'Data', 'Data']]
        expected_category_table = [['1.4', ['TiN'], ['Tc']], ['y', ['TiN'], ['A']], ['1.1', ['TiC'], ['Tc']], ['x', ['TiC'], ['A']], ['9.2', ['TaC'], ['Tc']], ['x', ['TaC'], ['A']], ['10.1', ['NbC'], ['Tc']], ['a', ['NbC'], ['A']], ['2.82', ['ZrB'], ['Tc']], ['x', ['ZrB'], ['A']], ['4.2', ['TaSi'], ['Tc']], ['x', ['TaSi'], ['A']], ['4.1', ['PbS'], ['Tc']], ['x', ['PbS'], ['A']], ['8.4', ['Pb-As alloy'], ['Tc']], ['x', ['Pb-As alloy'], ['A']], ['8.5', ['Pb-Sn-Bi'], ['Tc']], ['x', ['Pb-Sn-Bi'], ['A']], ['9.0', ['Pb-As-Bi'], ['Tc']], ['x', ['Pb-As-Bi'], ['A']], ['8.9', ['Pb-Bi-Sb'], ['Tc']], ['x', ['Pb-Bi-Sb'], ['A']]]
        self.do_subtable(input_path, 1, expected_labels, expected_category_table)
        expected_labels = [['StubHeader', 'ColHeader'], ['RowHeader', 'Data'], ['RowHeader', 'Data'], ['RowHeader', 'Data'], ['RowHeader', 'Data'], ['RowHeader', 'Data'], ['RowHeader', 'Data'], ['RowHeader', 'Data'], ['RowHeader', 'Data'], ['RowHeader', 'Data'], ['RowHeader', 'Data'], ['RowHeader', 'Data']]
        expected_category_table = [['1.1', ['TiO2'], ['Tc']], ['1.2', ['TiO3'], ['Tc']], ['1.3', ['TiO4'], ['Tc']], ['1.4', ['TiO5'], ['Tc']], ['1.5', ['TiO6'], ['Tc']], ['1.6', ['TiO7'], ['Tc']], ['1.7', ['TiO8'], ['Tc']], ['1.8', ['TiO9'], ['Tc']], ['1.9', ['TiO10'], ['Tc']], ['1.10', ['TiO11'], ['Tc']], ['1.11', ['TiO12'], ['Tc']]]
        self.do_subtable(input_path, 2, expected_labels, expected_category_table)

    def test_16(self):
        input_path = './tests/data/te_08.csv'
        expected_labels = [['StubHeader', 'StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'], ['RowHeader', 'RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data']]
        expected_category_table = [['2.431', ['La0.67Ba0.1Ca0.23MnO3', '1'], ['−ΔSM (J kg−1 K−1)']], ['42.37', ['La0.67Ba0.1Ca0.23MnO3', '1'], ['δTFWHM (K)']], ['103.001', ['La0.67Ba0.1Ca0.23MnO3', '1'], ['RCP (J kg−1)']], ['25.41', ['La0.67Ba0.1Ca0.23MnO3', '1'], ['ΔCp H max (J kg−1 K−1)']], ['−23.90', ['La0.67Ba0.1Ca0.23MnO3', '1'], ['ΔCp H min (J kg−1 K−1)']], ['Present', ['La0.67Ba0.1Ca0.23MnO3', '1'], ['Ref.']], ['3.721', ['La0.67Ba0.1Ca0.23MnO3', '2'], ['−ΔSM (J kg−1 K−1)']], ['40.16', ['La0.67Ba0.1Ca0.23MnO3', '2'], ['δTFWHM (K)']], ['149.435', ['La0.67Ba0.1Ca0.23MnO3', '2'], ['RCP (J kg−1)']], ['27.02', ['La0.67Ba0.1Ca0.23MnO3', '2'], ['ΔCp H max (J kg−1 K−1)']], ['−24.90', ['La0.67Ba0.1Ca0.23MnO3', '2'], ['ΔCp H min (J kg−1 K−1)']], ['Present', ['La0.67Ba0.1Ca0.23MnO3', '2'], ['Ref.']], ['4.736', ['La0.67Ba0.1Ca0.23MnO3', '3'], ['−ΔSM (J kg−1 K−1)']], ['40.75', ['La0.67Ba0.1Ca0.23MnO3', '3'], ['δTFWHM (K)']], ['192.992', ['La0.67Ba0.1Ca0.23MnO3', '3'], ['RCP (J kg−1)']], ['28.02', ['La0.67Ba0.1Ca0.23MnO3', '3'], ['ΔCp H max (J kg−1 K−1)']], ['−24.23', ['La0.67Ba0.1Ca0.23MnO3', '3'], ['ΔCp H min (J kg−1 K−1)']], ['Present', ['La0.67Ba0.1Ca0.23MnO3', '3'], ['Ref.']], ['5.742', ['La0.67Ba0.1Ca0.23MnO3', '4'], ['−ΔSM (J kg−1 K−1)']], ['42.32', ['La0.67Ba0.1Ca0.23MnO3', '4'], ['δTFWHM (K)']], ['243.001', ['La0.67Ba0.1Ca0.23MnO3', '4'], ['RCP (J kg−1)']], ['28.19', ['La0.67Ba0.1Ca0.23MnO3', '4'], ['ΔCp H max (J kg−1 K−1)']], ['−23.88', ['La0.67Ba0.1Ca0.23MnO3', '4'], ['ΔCp H min (J kg−1 K−1)']], ['Present', ['La0.67Ba0.1Ca0.23MnO3', '4'], ['Ref.']], ['5.971', ['La0.67Ba0.1Ca0.23MnO3', '5'], ['−ΔSM (J kg−1 K−1)']], ['46.65', ['La0.67Ba0.1Ca0.23MnO3', '5'], ['δTFWHM (K)']], ['278.547', ['La0.67Ba0.1Ca0.23MnO3', '5'], ['RCP (J kg−1)']], ['28.70', ['La0.67Ba0.1Ca0.23MnO3', '5'], ['ΔCp H max (J kg−1 K−1)']], ['−24.39', ['La0.67Ba0.1Ca0.23MnO3', '5'], ['ΔCp H min (J kg−1 K−1)']], ['Present', ['La0.67Ba0.1Ca0.23MnO3', '5'], ['Ref.']], ['2.2', ['La0.7Ca0.3MnO3', '2'], ['−ΔSM (J kg−1 K−1)']], ['NoValue', ['La0.7Ca0.3MnO3', '2'], ['δTFWHM (K)']], ['55', ['La0.7Ca0.3MnO3', '2'], ['RCP (J kg−1)']], ['NoValue', ['La0.7Ca0.3MnO3', '2'], ['ΔCp H max (J kg−1 K−1)']], ['NoValue', ['La0.7Ca0.3MnO3', '2'], ['ΔCp H min (J kg−1 K−1)']], ['19', ['La0.7Ca0.3MnO3', '2'], ['Ref.']], ['2.26', ['La0.6Sr0.2Ba0.2MnO3', '1'], ['−ΔSM (J kg−1 K−1)']], ['NoValue', ['La0.6Sr0.2Ba0.2MnO3', '1'], ['δTFWHM (K)']], ['67', ['La0.6Sr0.2Ba0.2MnO3', '1'], ['RCP (J kg−1)']], ['NoValue', ['La0.6Sr0.2Ba0.2MnO3', '1'], ['ΔCp H max (J kg−1 K−1)']], ['NoValue', ['La0.6Sr0.2Ba0.2MnO3', '1'], ['ΔCp H min (J kg−1 K−1)']], ['34', ['La0.6Sr0.2Ba0.2MnO3', '1'], ['Ref.']], ['1.46', ['La0.60Y0.07Ca0.33MnO3', '3'], ['−ΔSM (J kg−1 K−1)']], ['NoValue', ['La0.60Y0.07Ca0.33MnO3', '3'], ['δTFWHM (K)']], ['140', ['La0.60Y0.07Ca0.33MnO3', '3'], ['RCP (J kg−1)']], ['NoValue', ['La0.60Y0.07Ca0.33MnO3', '3'], ['ΔCp H max (J kg−1 K−1)']], ['NoValue', ['La0.60Y0.07Ca0.33MnO3', '3'], ['ΔCp H min (J kg−1 K−1)']], ['35', ['La0.60Y0.07Ca0.33MnO3', '3'], ['Ref.']], ['4', ['La0.7Sr0.3Mn0.93Fe0.07O3', '5'], ['−ΔSM (J kg−1 K−1)']], ['NoValue', ['La0.7Sr0.3Mn0.93Fe0.07O3', '5'], ['δTFWHM (K)']], ['255', ['La0.7Sr0.3Mn0.93Fe0.07O3', '5'], ['RCP (J kg−1)']], ['NoValue', ['La0.7Sr0.3Mn0.93Fe0.07O3', '5'], ['ΔCp H max (J kg−1 K−1)']], ['NoValue', ['La0.7Sr0.3Mn0.93Fe0.07O3', '5'], ['ΔCp H min (J kg−1 K−1)']], ['36', ['La0.7Sr0.3Mn0.93Fe0.07O3', '5'], ['Ref.']], ['2.68', ['La0.67Sr0.33MnO3', '2'], ['−ΔSM (J kg−1 K−1)']], ['NoValue', ['La0.67Sr0.33MnO3', '2'], ['δTFWHM (K)']], ['85', ['La0.67Sr0.33MnO3', '2'], ['RCP (J kg−1)']], ['NoValue', ['La0.67Sr0.33MnO3', '2'], ['ΔCp H max (J kg−1 K−1)']], ['NoValue', ['La0.67Sr0.33MnO3', '2'], ['ΔCp H min (J kg−1 K−1)']], ['37', ['La0.67Sr0.33MnO3', '2'], ['Ref.']]]
        self.do_table(input_path, expected_labels, expected_category_table)


if __name__ == '__main__':
    unittest.main()




# -*- coding: utf-8 -*-
"""
Test Footnote() class on some input tables
These tests depend on some of the features of Table() working properly.
Juraj Mavračić (jm2111@cam.ac.uk)
"""

import unittest
import logging

from tabledataextractor import Table
import numpy as np

log = logging.getLogger(__name__)


class TableF(Table):
    """Derivative of Table used to isolate testing for Footnotes()"""

    def __init__(self, file_path, table_number=1, **kwargs):
        super().__init__(file_path, table_number, **kwargs)

    @property
    def labels(self):
        """Cell labels. Python List"""
        temp = np.empty_like(self._pre_cleaned_table, dtype="<U60")
        temp[:, :] = '/'
        temp[self._cc4] = 'CC4'

        for footnote in self.footnotes:
            temp[footnote.prefix_cell[0], footnote.prefix_cell[1]] = 'FNprefix'
            if footnote.text_cell is not None:
                temp[footnote.text_cell[0], footnote.text_cell[1]] = 'FNtext' if temp[footnote.text_cell[0], footnote.text_cell[1]] == '/' else 'FNprefix & FNtext'
            for ref_cell in footnote.reference_cells:
                temp[ref_cell[0], ref_cell[1]] = 'FNref' if temp[ref_cell[0], ref_cell[1]] == '/' else temp[ref_cell[0], ref_cell[1]] + ' & FNref'

        return temp


class TestFootnote(unittest.TestCase):

    def test_table_use_footnotes(self):
        table = TableF("./tests/data/table_example_footnotes.csv", use_footnotes=True, use_spanning_cells=False)
        table.print()

        print(table.footnotes[0])
        self.footnote_1(table.footnotes[0])
        print(table.footnotes[1])
        self.footnote_2(table.footnotes[1])
        print(table.footnotes[2])
        self.footnote_3(table.footnotes[2])
        print(table.footnotes[3])
        self.footnote_4(table.footnotes[3])
        print(table.footnotes[4])
        self.footnote_5(table.footnotes[4])

        self.table_update(table.pre_cleaned_table.tolist())

    def footnote_1(self, fn):
        self.assertEqual("c", fn.prefix)
        self.assertEqual("Footnote text.", fn.text)
        self.assertListEqual(['OECD/DAC c'], fn.references)

    def footnote_2(self, fn):
        self.assertEqual("*", fn.prefix)
        self.assertEqual("Test", fn.text)
        self.assertListEqual(['2010*', '2011* a.'], fn.references)

    def footnote_3(self, fn):
        self.assertEqual("†)", fn.prefix)
        self.assertEqual("Source: OECD.", fn.text)
        self.assertListEqual(['2010†)', '2011†)'], fn.references)

    def footnote_4(self, fn):
        self.assertEqual("2", fn.prefix)
        self.assertEqual("", fn.text)
        self.assertListEqual(['New Zealand 2'], fn.references)

    def footnote_5(self, fn):
        self.assertEqual("a.", fn.prefix)
        self.assertEqual("whataboutthis", fn.text)
        self.assertListEqual(['2011 Test  a.', 'a.'], fn.references)

    def table_update(self, pre_cleaned_table):
        """Tests if the table has been correctly updated with the footnotes"""
        expected = [['1 Official development assistance', '', '', '', '', '', ''],
                    ['Country', 'Million dollar', 'Million dollar', 'Million dollar', 'Percentage of GNI', 'Percentage of GNI', 'Percentage of GNI'],
                    ['', '2007', '2010 Test ', '2011 Test   whataboutthis ', '2007', '2010 Source: OECD. ', '2011 Source: OECD. '],
                    [' whataboutthis ', '3735', '4580', '4936', '0.95', '1.1', '1'],
                    ['2', '2669', '3826', '4799', '0.32', '0.32', '0.35'],
                    ['New Zealand ', '320', '342', '429', '0.27', '0.26', '0.28'],
                    ['OECD/DAC Footnote text.', '104206', '128465', '133526', '0.27', '0.32', '0.31'],
                    ['c', 'Footnote text.', '', '', '', '', ''],
                    ['* Test', 'This is now just a note', ' because the footnote text was found on the left', '', '', '', ''],
                    ['†) Source: OECD.', '', '', '', '', '', ''],
                    ['2', '', '', '', '', '', ''],
                    ['a.whataboutthis', '', '', '', '', '', ''],
                    ['0.32 This should not be recognized as a footnote', 'This should not be recognized as a footnote.', '', '', '', '', '']]
        self.assertListEqual(expected, pre_cleaned_table)

    def test_table_dont_use_footnotes(self):
        table = TableF("./tests/data/table_example_footnotes.csv", use_footnotes=False, use_spanning_cells=False)
        table.print()

        print(table.footnotes[0])
        self.footnote_1b(table.footnotes[0])
        print(table.footnotes[1])
        self.footnote_2b(table.footnotes[1])
        print(table.footnotes[2])
        self.footnote_3b(table.footnotes[2])
        print(table.footnotes[3])
        self.footnote_4b(table.footnotes[3])
        print(table.footnotes[4])
        self.footnote_5b(table.footnotes[4])

        self.table_no_update(table.pre_cleaned_table.tolist())

    def footnote_1b(self, fn):
        self.assertEqual("c", fn.prefix)
        self.assertEqual("Footnote text.", fn.text)
        self.assertListEqual(['OECD/DAC c'], fn.references)

    def footnote_2b(self, fn):
        self.assertEqual("*", fn.prefix)
        self.assertEqual("Test", fn.text)
        self.assertListEqual(['2010*', '2011* a.'], fn.references)

    def footnote_3b(self, fn):
        self.assertEqual("†)", fn.prefix)
        self.assertEqual("Source: OECD.", fn.text)
        self.assertListEqual(['2010†)', '2011†)'], fn.references)

    def footnote_4b(self, fn):
        self.assertEqual("2", fn.prefix)
        self.assertEqual("", fn.text)
        self.assertListEqual(['New Zealand 2'], fn.references)

    def footnote_5b(self, fn):
        self.assertEqual("a.", fn.prefix)
        self.assertEqual("whataboutthis", fn.text)
        self.assertListEqual(['2011* a.', 'a.'], fn.references)

    def table_no_update(self, pre_cleaned_table):
        """Tests if the table has been correctly updated with the footnotes"""
        expected = [['1 Official development assistance', '', '', '', '', '', ''],
                    ['Country', 'Million dollar', 'Million dollar', 'Million dollar', 'Percentage of GNI', 'Percentage of GNI', 'Percentage of GNI'],
                    ['', '2007', '2010*', '2011* a.', '2007', '2010†)', '2011†)'],
                    ['a.', '3735', '4580', '4936', '0.95', '1.1', '1'],
                    ['2', '2669', '3826', '4799', '0.32', '0.32', '0.35'],
                    ['New Zealand 2', '320', '342', '429', '0.27', '0.26', '0.28'],
                    ['OECD/DAC c', '104206', '128465', '133526', '0.27', '0.32', '0.31'],
                    ['c', 'Footnote text.', '', '', '', '', ''],
                    ['* Test', 'This is now just a note', ' because the footnote text was found on the left', '', '', '', ''],
                    ['†) Source: OECD.', '', '', '', '', '', ''],
                    ['2', '', '', '', '', '', ''],
                    ['a.whataboutthis', '', '', '', '', '', ''],
                    ['0.32 This should not be recognized as a footnote', 'This should not be recognized as a footnote.', '', '', '', '', '']]
        self.assertListEqual(expected, pre_cleaned_table)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
Tests the downwards header extension

.. codeauthor: Juraj Mavračić <jm2111@cam.ac.uk>
"""

import unittest
import logging

from tabledataextractor import Table

log = logging.getLogger(__name__)


class TestHeaderExtDown(unittest.TestCase):

    def test_rows_extension(self):
        table = Table('./tests/data/te_03.csv')
        table.print()
        category_table = [['4.64', ['This study'], ['Test2', 'A']], ['2.99', ['This study'], ['Test3', 'B']], ['0.305', ['This study'], ['Test4', 'C']], ['3.83', ['This study'], ['Test5', 'D']], ['9.62', ['This study'], ['Test6', 'E']], ['0.208', ['This study'], ['Test7', 'F']]]
        labels = [['TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle'], ['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'], ['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data']]
        self.assertListEqual(category_table, table.category_table)
        self.assertListEqual(labels, table.labels.tolist())

    def test_no_rows_extension(self):
        table = Table('./tests/data/table_example5.csv')
        table.transpose()
        table.print()
        category_table = [['24.3', ['Short messages/thousands', 'Change %'], ['2003', '1647218']], ['33.2', ['Short messages/thousands', 'Change %'], ['2004', '2193498']], ['347', ['Short messages/thousands', ' Short messages/subscription'], ['2003', '1647218']], ['439', ['Short messages/thousands', ' Short messages/subscription'], ['2004', '2193498']], ['2314', ['Short messages/thousands', 'Multimedia messages/thousands'], ['2003', '1647218']], ['7386', ['Short messages/thousands', 'Multimedia messages/thousands'], ['2004', '2193498']], ['NoValue', ['Multimedia messages/thousands', 'Change %'], ['2003', '1647218']], ['219.2', ['Multimedia messages/thousands', 'Change %'], ['2004', '2193498']]]
        labels = [['StubHeader', 'StubHeader', 'ColHeader', 'ColHeader'], ['StubHeader', 'StubHeader', 'ColHeader', 'ColHeader'], ['RowHeader', 'RowHeader', 'Data', 'Data'], ['RowHeader', 'RowHeader', 'Data', 'Data'], ['RowHeader', 'RowHeader', 'Data', 'Data'], ['RowHeader', 'RowHeader', 'Data', 'Data']]
        self.assertListEqual(category_table, table.category_table)
        self.assertListEqual(labels, table.labels.tolist())

    def test_col_extension(self):
        table = Table('./tests/data/te_03.csv')
        table.transpose()
        table.print()
        category_table = [['4.64', ['Test', 'Test2', 'A'], ['This study']], ['2.99', ['Test', 'Test3', 'B'], ['This study']], ['0.305', ['Test', 'Test4', 'C'], ['This study']], ['3.83', ['Test', 'Test5', 'D'], ['This study']], ['9.62', ['Test', 'Test6', 'E'], ['This study']], ['0.208', ['Test', 'Test7', 'F'], ['This study']]]
        labels = [['StubHeader', 'StubHeader', 'StubHeader', 'ColHeader'], ['RowHeader', 'RowHeader', 'RowHeader', 'Data'], ['RowHeader', 'RowHeader', 'RowHeader', 'Data'], ['RowHeader', 'RowHeader', 'RowHeader', 'Data'], ['RowHeader', 'RowHeader', 'RowHeader', 'Data'], ['RowHeader', 'RowHeader', 'RowHeader', 'Data'], ['RowHeader', 'RowHeader', 'RowHeader', 'Data']]
        self.assertListEqual(category_table, table.category_table)
        self.assertListEqual(labels, table.labels.tolist())

    def test_no_col_extension(self):
        table = Table('./tests/data/table_example5.csv')
        table.print()
        category_table = [['1647218', ['2003'], ['', 'Short messages/thousands']], ['24.3', ['2003'], ['Short messages/thousands', 'Change %']], ['347', ['2003'], ['Short messages/thousands', ' Short messages/subscription']], ['2314', ['2003'], ['Short messages/thousands', 'Multimedia messages/thousands']], ['NoValue', ['2003'], ['Multimedia messages/thousands', 'Change %']], ['2193498', ['2004'], ['', 'Short messages/thousands']], ['33.2', ['2004'], ['Short messages/thousands', 'Change %']], ['439', ['2004'], ['Short messages/thousands', ' Short messages/subscription']], ['7386', ['2004'], ['Short messages/thousands', 'Multimedia messages/thousands']], ['219.2', ['2004'], ['Multimedia messages/thousands', 'Change %']]]
        labels = [['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'], ['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data']]
        self.assertListEqual(category_table, table.category_table)
        self.assertListEqual(labels, table.labels.tolist())


if __name__ == '__main__':
    unittest.main()
